"""
Static values of the codec
"""


class SubframeType:
    CONSTANT = 0b00
    RAW = 0b01
    LPC = 0b10
    LPC_COMMON = 0b11


col_types = {
    "seq": int,  # int - frame number, if grouped by this then represents a slice across all channels
    "frame": object,  # np.array - raw audio samples (and warmup samples when decoding)
    "channel": int,  # int - channel number or also subframe number
    "qlp": object,  # np.array - quantized lpc coefficients -> len() is the order
    "qlp_precision": float,  # float (because of nans...) - precision of stored LPC coefficients in bits
    "shift": float,  # float (because of nans...) - shift needed (in bits) to revert the quantization
    "residual": object,  # pd.array - the signal residual
    "bps": int,  # int - bits per sample or also the k parameter of rice coding
    "frame_type": int,  # int (enum) - corresponds to one of SUBFRAME_CONSTANT, SUBFRAME_RAW, SUBFRAME_LPC
}

columns = list(col_types.keys())

soundfile_dtype = {
    16: 16,
    24: 32,
    32: 32,
}


class Default:
    min_frame_size = 1 << 11
    max_frame_size = 1 << 13
    framing_treshold = 20000
    framing_resolution = 10
    rice_responsiveness = 20



# distutils: language = c
# cython: language_level=3
import cython
import numpy as np

####################
# LPC coefficients #
####################

# @cython.cdivision(True)
# def compute_lp_coefficients(long[:] autoc, int order, double[:] lpc):
#     cdef int j
#     cdef double r, err
#     cdef double tmp
#
#     err = autoc[0]
#
#     r = -autoc[order + 1]
#     for j in range(order):
#         r -= lpc[j] * autoc[order - j]
#     r /= err
#
#     lpc[order] = r
#     for j in range(order >> 1):
#         tmp = lpc[j]
#         lpc[j] += r * lpc[order - 1 - j]
#         lpc[order - 1 - j] += r * tmp
#
#     if order & 1:
#         lpc[j] += lpc[j] * r
#
#     err *= (1.0 - r * r)
#     return err


################
# Quantization #
################

def quantize_lpc(double[:] lpc_c, int precision) -> int:
    """
    Quantizes LPC coefficients to a given precision
    Implemented from: https://github.com/xiph/flac/blob/master/src/libFLAC/lpc.c
    :param lpc_c: numpy array of LPC coefficients to be quantized
    :param precision: target precition in bits
    :return: shift in bits
    """
    # reserve 1 bit for sign
    precision -= 1

    cdef int qmax = 1 << precision
    cdef int qmin = -qmax
    qmax -= 1

    cdef double cmax = np.max(np.abs(lpc_c))

    cdef int shift = precision - np.frexp(cmax)[1]

    # TODO: this should probably be reimplemented from FLAC
    # if shift > max_shiftlimit:
    #     shift = max_shiftlimit
    # elif shift < min_shiftlimit:
    #     return None, 0

    cdef int order = lpc_c.shape[0]
    cdef double error = 0.0
    cdef double q
    cdef int i
    for i in range(order):
        error += lpc_c[i] * (1 << shift)
        q = round(error)

        if q > qmax:
            q = qmax
        elif q < qmin:
            q = qmin

        error -= q
        lpc_c[i] = q

    return shift

##############
# Prediction #
##############

def compute_residual(cython.integral[:] frame, cython.integral[:] residual, int[:] qlp, int lp_quantization):
    cdef Py_ssize_t data_len = frame.shape[0]
    cdef Py_ssize_t order = qlp.shape[0]
    cdef Py_ssize_t  i, j
    cdef long _sum

    for i in range(order, data_len):
        _sum = 0
        for j in range(order):
            _sum += qlp[j] * frame[i - j - 1]
        residual[i] = frame[i] - (_sum >> lp_quantization)

###############
# Restoration #
###############


def restore_signal(cython.integral[:] frame, int[:] qlp, int lp_quantization):
    cdef Py_ssize_t data_len = frame.shape[0]
    cdef Py_ssize_t order = qlp.shape[0]
    cdef Py_ssize_t  i, j
    cdef long _sum

    if order <= 0:
        return None

    for i in range(order, data_len):
        _sum = 0
        for j in range(order):
            _sum += qlp[j] * frame[i - j - 1]
        frame[i] = frame[i] + (_sum >> lp_quantization)



import numpy as np

from straw.io.params import StreamParams
from .bias import BiasCorrector
from .decorrelator import Decorrelator
from .gain import GainCorrector
from .shift import ShiftCorrector


def apply_corrections(data: np.array,
                      corrections: tuple,
                      target_params: StreamParams = StreamParams(),
                      force_inplace: bool = False):
    """
    Apply the given corrections to the data in thegiven order
    :param data: array to apply the corrections to - ndarray with dimensions (channels, samples)
    :param corrections: corrections to apply
    :param target_params: stream params where the correction params should be stored
    :param force_inplace: force the operation inplace (mainly shift correction)
    :return:
    """
    for correction in corrections:
        if correction == "gain":
            GainCorrector().apply(data, target_params)
        elif correction == "bias":
            BiasCorrector().apply(data, target_params)
        elif correction == "shift":
            sc = ShiftCorrector()
            sc.apply(data, target_params)
            if force_inplace:
                sc.apply_to_ndarray(data)
        else:
            raise ValueError(f"Invalid correction name: '{correction}', must be one of ('gain', 'bias', 'shift')")



import sys
import timeit
from pathlib import Path

import numpy as np

from straw import Encoder, Decoder


def read(file) -> (np.array, int):
    """
    Read a Straw file (compatibility function to soundfile.read)
    :param file: Input file
    :return: soundfile compatible array and samplerate
    """
    d = Decoder()
    d.load_file(Path(file))
    d.decode()
    return d.get_soundfile_compatible_array(), d.get_params().sample_rate


def write(file, data: np.array, samplerate: int):
    """
    Write a Straw file (compatibility function to soundfile.write)
    :param file: Output file
    :param data: Data to be written
    :param samplerate: Sample rate of the given data
    :return: None
    """
    e = Encoder()
    e.load_data(data, samplerate, data.dtype.itemsize * 8)
    e.encode()
    e.save_file(Path(file).open("wb"))


def _encode(args):
    """
    Internal encoder call from the executable script
    :param args: command line args (see straw/__main__.py)
    :return: None
    """
    e = Encoder(flac_mode=False,
                dynamic_blocksize=args.dynamic_blocksize,
                min_block_size=args.min_frame_size,
                max_block_size=args.max_frame_size,
                framing_treshold=args.framing_treshold,
                framing_resolution=args.framing_resolution,
                responsiveness=args.rice_responsiveness,
                parallelize=args.parallel,
                show_progress=False)

    if not args.silent:
        print(f"Loading file ... ", end="", file=sys.stderr)
    start = timeit.default_timer()
    e.load_file(Path(args.input_files[0]))
    stop = timeit.default_timer()
    if not args.silent:
        print(f"DONE in {stop - start:.3f} s", file=sys.stderr)

    if not args.silent:
        print(f"Encoding ... ", end="", file=sys.stderr)
    mid = timeit.default_timer()
    e.encode()
    stop = timeit.default_timer()
    if not args.silent:
        print(f"DONE in {stop - mid:.3f} s", file=sys.stderr)

    if not args.silent:
        print(f"Saving to file ... ", end="", file=sys.stderr)
    mid = timeit.default_timer()
    e.save_file(args.output_file)
    stop = timeit.default_timer()
    if not args.silent:
        print(f"DONE in {stop - mid:.3f} s", file=sys.stderr)

    if args.verbose and not args.silent:
        e.print_stats(args.output_file, stream=sys.stderr)

    if not args.silent:
        print(f"Total processing time: {stop - start:.3f} s", file=sys.stderr)


def _decode(args):
    """
    Internal decoder call from the executable script
    :param args: command line args (see straw/__main__.py)
    :return: None
    """
    d = Decoder(flac_mode=False, show_progress=False)

    if not args.silent:
        print(f"Loading file ... ", end="", file=sys.stderr)
    start = timeit.default_timer()
    d.load_file(Path(args.input_files[0]))
    stop = timeit.default_timer()
    if not args.silent:
        print(f"DONE in {stop - start:.3f} s", file=sys.stderr)

    if not args.silent:
        print(f"Decoding ... ", end="", file=sys.stderr)
    mid = timeit.default_timer()
    d.decode()
    stop = timeit.default_timer()
    if not args.silent:
        print(f"DONE in {stop - mid:.3f} s", file=sys.stderr)

    if not args.silent:
        print(f"Saving to file ... ", end="", file=sys.stderr)
    mid = timeit.default_timer()
    d.save_file(args.output_file)
    stop = timeit.default_timer()
    if not args.silent:
        print(f"DONE in {stop - mid:.3f} s", file=sys.stderr)

    # if Path(args.input_files[0]).stem == "1min":
    #     d.test()

    if not args.silent:
        print(f"Total processing time: {stop - start:.3f} s", file=sys.stderr)


def run(args):
    """
    Run the encoder/decoder
    :param args: command line args (see straw/__main__.py)
    :return: None
    """
    if args.verbose and not args.silent:
        print(args, file=sys.stderr)
    if not args.decode:
        _encode(args)
    else:
        _decode(args)



import numpy as np
import pandas as pd

from straw.static import SubframeType


class Modifiers:
    #############################
    # Subtraction decorrelation #
    #############################
    @staticmethod
    def indicator(x):
        """
        A cutoff indicator function
        :param x:
        :return:
        """
        return np.mean(np.abs(x))

    @staticmethod
    def localized_sub(x1: np.array, x2: np.array) -> bool:
        """
        Performs localized subtraction on array x1 with the reference array (x2)
        The process is as follows:
            1. A cutoff value is calculated from each given array equal to (x2.max() / (2^3))
            2. Every value below this treshold is saved as is
            3. For every value above this reshold the respective indices value for the reference array is subtracted
            4. The variance of this new array is compared to the original
            5. If the new variance is larger, then the original array is returned instead
        :param x1: array to be modified
        :param x2: reference array
        :return: True if x1 was modified, False otherwise
        """
        diff = x1 - x2
        if diff.any():
            if Modifiers.indicator(diff) < Modifiers.indicator(x1):
                x1[:] = diff
                return True
            else:
                return False
        else:
            return False

    @staticmethod
    def localized_add(x1: np.array, x2: np.array, was_coded: bool):
        """
        Reverse localized_sub
        :param x1: array to be modified
        :param x2: reference array
        :param was_coded: True if x1 was modified, False otherwise
        :return: None
        """
        if not was_coded:
            return

        x1[:] = x2 + x1

    ##########################
    # Mid-side decorrelation #
    ##########################

    @staticmethod
    def transform_midside(x1: np.array, x2: np.array):
        diff = x1 - x2
        mid = (x1 + x2) >> 1
        x1[:] = diff
        x2[:] = mid

    @staticmethod
    def transform_midside_reverse(x1: np.array, x2: np.array):
        diff = x1.copy()
        mid = x2.copy()
        x1[:] = mid + (diff >> 1) + (diff & 1)
        x2[:] = mid - (diff >> 1)


class Decorrelator:
    #############################
    # Subtraction decorrelation #
    #############################

    @staticmethod
    def localized_decorrelate_expander(df: pd.DataFrame, reference: np.ndarray, col_name: str):
        return Modifiers.localized_sub(df[col_name], x2=reference)

    def localized_decorrelate(self, df: pd.DataFrame, col_name: str = "residual"):
        """
        Performs localized decorrelation
        The reference channel is channel 0
        :param df: dataframe slice for one multichannel group
        :param col_name: the name of the column which should be decorrelated
        :return: df with new added columns
        """
        if col_name not in df.columns:
            raise ValueError(f"Column '{col_name}' not in dataframe")

        df["was_coded"] = df.apply(self.localized_decorrelate_expander,
                                   reference=df[col_name][df.index[0]],
                                   col_name=col_name,
                                   axis=1)

        return df

    @staticmethod
    def localized_decorrelate_revert_expander(df: pd.DataFrame, reference: np.ndarray, col_name: str):
        Modifiers.localized_add(df[col_name], x2=reference, was_coded=df["was_coded"])

    def localized_decorrelate_revert(self, df: pd.DataFrame, col_name: str = "residual"):
        """
        Reverts localized decorrelation
        The reference channel is channel 0
        For the process description see Decorrelator.localized_decorrelate
        :param df: dataframe slice for one multichannel group
        :param col_name: the name of the column which was decorrelated
        :return: df with new added columns
        """
        if col_name not in df.columns:
            raise ValueError(f"Column '{col_name}' not in dataframe")

        df.apply(self.localized_decorrelate_revert_expander,
                 reference=df[col_name][df.index[0]],
                 col_name=col_name,
                 axis=1)

        return df

    ##########################
    # Mid-side decorrelation #
    ##########################

    @staticmethod
    def _find_closest_lower_power_of_two(x):
        order = 1
        while order <= x:
            order <<= 1
        return order >> 1

    @staticmethod
    def midside_decorrelate(df: pd.DataFrame, col_name: str = "residual", iterated: bool = True):
        """
        Apply the mid-side decorrelation
        :param df: source DataFrame
        :param col_name: column that should be decorrelated
        :param iterated: whether the decorrelation should be performed in iterations
        :return: modified DataFrame
        """
        if col_name not in df.columns:
            raise ValueError(f"Column '{col_name}' not in dataframe")

        if not (df["frame_type"] == SubframeType.LPC_COMMON).all() or len(df) == 1:
            return df

        order = Decorrelator._find_closest_lower_power_of_two(len(df))
        if order != len(df):
            iterated = False

        if iterated:
            indices = np.arange(order).reshape((-1, 2))
            while order > 1:
                indices = np.rot90(indices).reshape(-1, 2)
                for idx1, idx2 in indices:
                    Modifiers.transform_midside(df.loc[df.index[idx1], col_name], x2=df.loc[df.index[idx2], col_name])
                order = order >> 1
        else:
            indices = np.arange((len(df) // 2) * 2).reshape((-1, 2))
            for idx1, idx2 in indices:
                Modifiers.transform_midside(df.loc[df.index[idx1], col_name], x2=df.loc[df.index[idx2], col_name])

        return df

    @staticmethod
    def midside_decorrelate_revert(df: pd.DataFrame, col_name: str = "residual", iterated: bool = True):
        """
        Apply the reverse of mid-side decorrelation
        :param df: source DataFrame
        :param col_name: column that should be decorrelated
        :param iterated: whether the decorrelation should be performed in iterations
        :return: modified DataFrame
        """
        if col_name not in df.columns:
            raise ValueError(f"Column '{col_name}' not in dataframe")

        if not (df["frame_type"] == SubframeType.LPC_COMMON).all():
            return df

        if iterated:
            order = Decorrelator._find_closest_lower_power_of_two(len(df))
            indices = np.arange(order).reshape((-1, 2))

            while order > 1:
                indices = np.rot90(indices).reshape(-1, 2)
                order = order >> 1

            order = Decorrelator._find_closest_lower_power_of_two(len(df))
            while order > 1:
                for idx1, idx2 in indices:
                    Modifiers.transform_midside_reverse(df.loc[df.index[idx1], col_name],
                                                        x2=df.loc[df.index[idx2], col_name])
                indices = np.rot90(indices.reshape(2, -1), k=-1)
                order = order >> 1
        else:
            indices = np.arange((len(df) // 2) * 2).reshape((-1, 2))
            for idx1, idx2 in indices:
                Modifiers.transform_midside_reverse(df.loc[df.index[idx1], col_name],
                                                    x2=df.loc[df.index[idx2], col_name])

        return df



from pathlib import Path

import soundfile

from straw import lpc
from straw.codec.base import BaseCoder
from straw.correctors import Decorrelator, GainCorrector, BiasCorrector
from straw.io import Formatter


class Decoder(BaseCoder):
    ##########
    # Public #
    ##########

    def load_file(self, input_file: Path):
        """
        Load the specified file into memory
        :param input_file: file to load
        :return: None
        """
        reader = Formatter().load(input_file, self._flac_mode, show_progress=self.show_progress)
        self._data = reader.get_data()
        self._params = reader.get_params()
        self._samplebuffer = reader.get_buffer()

    def decode(self):
        """
        Decode the signal and verify its integrity
        :return: None
        """
        self._revert_decorrelate("residual")
        self._data.apply(lpc.compute_original, axis=1, inplace=True)
        self._revert_corrections()
        if self.get_md5() != self._params.md5:
            print(f"md5 restored: {self.get_md5().hex(' ')}")
            print(f"md5 file:     {self._params.md5.hex(' ')}")
            raise ValueError("Non-matching md5 checksum")

    def save_file(self, output_file: Path):
        """
        Save the decoded signal
        :param output_file: target file
        :return: None
        """
        with soundfile.SoundFile(output_file, "w",
                                 subtype=self._subtype_pattern.format(self._params.bits_per_sample),
                                 samplerate=self._params.sample_rate,
                                 channels=self._params.channels) as wav:
            dtype_bits = self._samplebuffer.itemsize * 8
            shift = dtype_bits - self._params.bits_per_sample
            data = (self._samplebuffer.swapaxes(1, 0) << shift)
            wav.write(data)

    ###########
    # Private #
    ###########

    def _revert_corrections(self):
        BiasCorrector().apply_revert(self._samplebuffer, self._params)
        GainCorrector().apply_revert(self._samplebuffer, self._params)

    def _revert_decorrelate(self, col_name="residual"):
        self._data = self._data.groupby("seq").apply(Decorrelator().midside_decorrelate_revert, col_name=col_name)
        # self._data = self._data.groupby("seq").apply(Decorrelator().localized_decorrelate_revert, col_name=col_name)

    ###########
    # Utility #
    ###########

    def test(self, wav_file: Path = Path("inputs/1min.wav")):
        """
        Test the decoder with a sample file
        :return:
        """
        if not wav_file.exists():
            return

        tmp, _ = soundfile.read(wav_file, dtype="int16")
        tmp = tmp.swapaxes(1, 0)
        if not (self._samplebuffer - tmp).any():
            print("Lossless :)")
        else:
            print("Not lossless :|")



# distutils: language = c
# cython: language_level=3
import cython

###########
# Reading #
###########

@cython.cdivision(True)
def read_frame(cython.integral[:] target, const unsigned char[:] source, Py_ssize_t bits_per_sample, char little_endian = 0):
    cdef Py_ssize_t source_size, i, byteshift
    cdef Py_ssize_t bytes_per_sample = bits_per_sample / 8
    source_size = source.shape[0]

    for i in range(source_size):
        byteshift = i % bytes_per_sample
        if little_endian:
            byteshift = bytes_per_sample - byteshift - 1
        target[i / bytes_per_sample] |= source[i] << (8 * byteshift)

@cython.cdivision(True)
def write_frame(unsigned char[:] target, cython.integral[:] source, Py_ssize_t bits_per_sample, char little_endian = 0):
    cdef Py_ssize_t source_size, i, byte_i, byteshift
    cdef Py_ssize_t bytes_per_sample = bits_per_sample / 8
    source_size = source.shape[0]

    for i in range(source_size):
        for byte_i in range(bytes_per_sample):
            byteshift = byte_i
            if not little_endian:
                byteshift = bytes_per_sample - byteshift - 1

            target[bytes_per_sample*i + byte_i] = (source[i] >> (8 * byteshift)) & 0xff



import mmap
from pathlib import Path
from typing import BinaryIO

import numpy as np
import pandas as pd
from crcmod import mkCrcFun

from straw import static
from straw.io.bitarray import SlicedBitarray
from straw.io.params import StreamParams
from straw.rice import Ricer


class BaseIO:
    """
    Base class for Readers and Writers
    """
    _data: pd.DataFrame
    _params: StreamParams
    _f = None

    class Crc:
        crc8 = mkCrcFun(0x107, initCrc=0, rev=False)
        crc16 = mkCrcFun(0x18005, initCrc=0, rev=False)

    def _format_specific_checks(self):
        """
        This should be overridden
        :return: None
        """
        pass

    def _stream(self):
        """
        This should be overridden
        :return: None
        """
        pass

    def get_data(self):
        """
        Query the data view
        :return: data view
        """
        return self._data

    def get_params(self):
        """
        Query the params
        :return: params
        """
        return self._params


class BaseWriter(BaseIO):
    def __init__(self, initial_params: StreamParams, output_stream: BinaryIO):
        self._params = initial_params
        self._f = output_stream
        self._stream()

    def init_stream(self, params):
        self._params = params
        self._stream()

    def write(self, data):
        secs = data.groupby("seq").apply(self._frame)
        for sec in secs:
            sec.tofile(self._f)

    def close_stream(self, params):
        self._f.seek(0)
        self._params = params
        self._stream()

    def _frame(self, df: pd.DataFrame):
        """
        This should be overridden
        :return: None
        """
        pass

    @staticmethod
    def encode_int_utf8(val: int) -> bytes:
        return chr(val).encode("utf-8")


class BaseReader(BaseIO):
    _raw: list
    _ricer = Ricer()

    _sec: SlicedBitarray()
    _memview: bytes
    _samplebuffer_ptr: int = 0
    _samplebuffer: np.array

    def __init__(self, show_progress: bool = True):
        self._params = StreamParams()
        self._raw = []
        self.show_progress = show_progress

    def load(self, input_file: Path) -> (pd.DataFrame, StreamParams):
        """
        Saves the dataframe into a FLAC formatted binary file
        :param input_file: source file
        :return: dataframe and params
        """
        with input_file.open("rb") as f:
            m = mmap.mmap(f.fileno(), 0, access=mmap.ACCESS_READ)
            self._sec = SlicedBitarray(buffer=m)
            self._memview = memoryview(self._sec).tobytes()
            self._stream()
        self._format_specific_checks()
        self._raw.sort(key=lambda x: x["idx"])
        self._data = pd.DataFrame(self._raw, columns=static.columns, copy=False)

    def _allocate_buffer(self):
        """
        Allocate the samplebuffer
        :return: None
        """
        channels = self._params.channels
        total_samples = self._params.total_samples
        dtype_bits = static.soundfile_dtype[self._params.bits_per_sample]
        self._samplebuffer = np.zeros((total_samples, channels), dtype=f"int{dtype_bits}")
        self._samplebuffer = self._samplebuffer.swapaxes(1, 0)

    def get_buffer(self):
        """
        Query the samplebuffer view
        :return: samplebuffer view
        """
        return self._samplebuffer



import numpy as np

from straw.correctors.base import BaseCorrector
from straw.io.params import StreamParams


class BiasCorrector(BaseCorrector):
    def apply(self, samplebuffer: np.ndarray, params: StreamParams) -> (np.ndarray, np.ndarray):
        bias = np.zeros(samplebuffer.shape[0], dtype=np.int8)
        for i in range(samplebuffer.shape[0]):
            bias[i] = samplebuffer[i].mean()
            samplebuffer[i] -= bias[i]
        params.bias = bias

    def apply_revert(self, samplebuffer: np.ndarray, params: StreamParams) -> (np.ndarray, np.ndarray):
        for i in range(params.channels):
            samplebuffer[i] += params.bias[i]



import numpy as np

from straw.static import Default


def zero_crossings(x, zero_pos: bool = True):
    """
    Similar output to librosa.zero_crossings but more efficient
    :param x: signal
    :param zero_pos: whether to count the first number as a zero crossing
    :return: Zero crossings
    """
    if zero_pos:
        x_sign = np.signbit(x)
    else:
        x_sign = np.sign(x)

    return np.insert(x_sign[1:] != x_sign[:-1], 0, 1)


def resample_indices(raw_indices: np.array, min_block_size: int, max_block_size: int):
    """
    Resample the indices so that the new block sizes are between min_block_size and max_block_size
    :param raw_indices: indices to be resampled
    :param min_block_size: minimal allowed block size
    :param max_block_size: maximal allowed block size
    :return: new indices
    """
    indices = [0]
    currect_size = 0
    for i in range(1, raw_indices.shape[0]):
        currect_size += raw_indices[i] - raw_indices[i - 1]
        if min_block_size <= currect_size <= max_block_size:
            indices.append(raw_indices[i])
            currect_size = 0

        while currect_size > max_block_size:
            indices.append(indices[-1] + max_block_size)
            currect_size -= max_block_size

    indices[-1] = raw_indices[-1]
    return np.asarray(indices)


class Signals:
    @classmethod
    def get_frame_limits_by_energy(cls,
                                   channel_data: np.array,
                                   min_block_size: int = Default.min_frame_size,
                                   treshold: int = Default.framing_treshold,
                                   max_block_size: int = Default.max_frame_size,
                                   resolution: int = Default.framing_resolution):
        """
        Returns a list of indices where frames should start, determined by the short-time energy
        :param channel_data: full data of one channel
        :param min_block_size: minimal allowed block size
        :param treshold: energy threshold
        :param max_block_size: maximal allowed block size
        :param resolution: framing resolution
        :return: Indices of frame borders including the last border
        """
        if resolution is None:
            resolution = min_block_size
        energies = cls.get_energies(channel_data, resolution, treshold)
        # Find indices where energy crosses a treshold
        # a crossing up means a high energy frame, crossing down a low energy frame
        # borders = zero_crossings(energies)
        borders = zero_crossings(energies - treshold, zero_pos=False)
        all_indices = borders.nonzero()[0] * resolution
        all_indices[-1] = channel_data.shape[0]

        indices = resample_indices(all_indices, min_block_size, max_block_size)

        return indices

    @staticmethod
    def get_energies(data: np.array,
                     resolution: int = Default.framing_resolution,
                     treshold: int = Default.framing_treshold):
        """
        Return the energies of a frame
        :param data: frame data
        :param resolution: framing resolution
        :param treshold: energy threshold
        :return: short-term energies in a frame
        """
        data = data.astype(np.int64)
        lst = []
        for i in range(0, data.shape[0], resolution):
            fr = data[i:i + resolution]
            # shorttime_energy = Signals.signaltonoise(fr)
            shorttime_energy = np.sum(fr * fr) / fr.shape[0]
            lst.append(shorttime_energy)
        lst.append(treshold)
        return np.asarray(lst)



import numpy as np

from straw.correctors.base import BaseCorrector
from straw.io.params import StreamParams
from straw.io.sizes import StrawSizes


class GainCorrector(BaseCorrector):
    def apply(self, samplebuffer: np.ndarray, params: StreamParams) -> (np.ndarray, np.ndarray):
        factors = self.find_factors(samplebuffer, StrawSizes.metadata_block_streaminfo.gain)
        for i in range(samplebuffer.shape[0]):
            # scaling by 1.0 is just a waste of time
            if factors[i] == 1.0:
                continue

            self.equalize(samplebuffer[i], factors[i])
        params.gain, params.gain_shift = self.quantize_factors(factors, StrawSizes.metadata_block_streaminfo.gain)

    def apply_revert(self, samplebuffer: np.ndarray, params: StreamParams) -> (np.ndarray, np.ndarray):
        if params.gain_shift is None:
            return

        factors = self.dequantize_factors(params.gain, params.gain_shift)
        for i in range(samplebuffer.shape[0]):
            # scaling by 1.0 is just a waste of time
            if factors[i] == 1.0:
                continue

            self.deequalize(samplebuffer[i], factors[i])

    @staticmethod
    def quantize_factors(factors: np.ndarray, precision: int):
        """
        Quantize the factors to integer values
        :param factors: numpy array of factors in floating point format
        :param precision: desired quantization precision in bits
        :return: numpy array of factors in integer format and shift
        """
        factors -= 1.0
        qmax = 1 << precision
        qmin = -qmax
        qmax -= 1

        for shift in reversed(range(0, precision + 1)):
            vals = (factors * (1 << shift)).astype(np.int64)
            if vals.max() <= qmax and vals.min() >= qmin:
                return vals, shift

    @staticmethod
    def dequantize_factors(factors: np.ndarray, shift: int):
        """
        Dequantizes the factors to a floating point format
        :param factors: numpy array of factors in integer format
        :param shift: shift performed at quantization
        :return: numpy array of factors in floating point format
        """
        return (factors / (1 << shift)) + 1.0

    @staticmethod
    def energy(frame: np.ndarray):
        """
        Return the normalized energy of a frame
        :param frame: input frame
        :return: normalized energy
        """
        return frame.std()

    @staticmethod
    def equalize(frame: np.ndarray, factor: float):
        """
        Scale a channel with the given factor
        NOTE: this is performed inplace
        :param frame: input frame
        :param factor: the scaling factor
        :return: None
        """
        frame[:] = (frame * factor).round().astype(frame.dtype)

    @staticmethod
    def deequalize(frame: np.ndarray, factor: float):
        """
        Reverse the scaling of a channel with the given factor
        NOTE: this is performed inplace
        :param frame: input frame
        :param factor: the scaling factor
        :return: None
        """
        frame[:] = (frame / factor).round().astype(frame.dtype)

    def find_factors(self, samplebuffer: np.ndarray, precision: int):
        """
        Find the factors with which the channels should be multiplied
        :param samplebuffer: array to apply the corrections to - ndarray with dimensions (channels, samples)
        :param precision: desired quantization precision in bits
        :return:
        """
        energies = np.asarray([self.energy(channel_data) for channel_data in samplebuffer])
        strongest_idx = energies.argmax()
        factors = energies[strongest_idx] / energies

        # Apply and de-apply quantization
        factors, shift = self.quantize_factors(factors, precision)
        factors = self.dequantize_factors(factors, shift)
        return factors



import numpy as np
import pandas as pd
from scipy.signal import get_window

from straw.compute import ParallelCompute
from straw.correctors.base import BaseCorrector
from straw.io.params import StreamParams


class ShiftCorrector(BaseCorrector):
    _lags: np.array = None

    def apply(self, samplebuffer: np.ndarray, params: StreamParams, limit=10, parallel=True) -> (
    np.ndarray, np.ndarray):
        windowed = (samplebuffer * get_window("nuttall", samplebuffer.shape[1])).astype(np.int64)
        leading_channel = self._find_leading_channel(windowed, limit=limit, parallel=parallel)
        params.lags = self._find_lags(windowed, leading_channel, limit=limit)
        params.leading_channel = leading_channel
        total_size = samplebuffer.shape[1] - np.max(params.lags)
        for i in range(samplebuffer.shape[0]):
            lag = params.lags[i]
            params.removed_samples_start.append(samplebuffer[i][:lag])
            params.removed_samples_end.append(samplebuffer[i][total_size + lag:])
        self._lags = params.lags

    def apply_to_ndarray(self, data):
        """
        Performs shift inplace by rewriting each sample by a rolled array
        NOTE: this is very inefficient and this operation is normally performed at framing/blocking
         and this method is used only in figures where it needs to be done explicitly
        """
        for channel in range(data.shape[0]):
            lag = self._lags[channel]
            data[channel] = np.roll(data[channel], -lag)

    def df_wrap_apply(self, frameset: pd.Series):
        ndarr = np.stack(frameset.tolist())
        self.apply(ndarr, StreamParams(), parallel=False)
        self.apply_to_ndarray(ndarr)
        for i, idx in enumerate(frameset.index):
            frameset[idx][:] = ndarr[i]

    def _find_leading_channel(self, samplebuffer: np.ndarray, limit, parallel):
        lags = np.zeros(samplebuffer.shape[0], dtype=np.int8)
        reference = samplebuffer[0]
        if parallel:
            lags[:] = ParallelCompute.get_instance().map_ndarray(samplebuffer, self._double_sided_corr,
                                                                 reference=reference,
                                                                 limit=limit)
        else:
            for i in range(1, samplebuffer.shape[0]):
                lags[i] = self._double_sided_corr(samplebuffer[i], reference=reference, limit=limit)

        return np.argmin(lags)

    def _find_lags(self, samplebuffer: np.ndarray, leading_channel, limit) -> np.ndarray:
        lags = np.zeros(samplebuffer.shape[0], dtype=np.int8)
        reference = samplebuffer[leading_channel]
        # lags[:] = ParallelCompute.get_instance().map_ndarray(samplebuffer, self._corr, reference=reference,
        #                                                      end=limit-1)
        # lags[leading_channel] = 0
        for i in range(samplebuffer.shape[0]):
            if i == leading_channel:
                continue
            lags[i] = self._corr(frame=samplebuffer[i], reference=reference, end=limit - 1)
        return lags

    def _double_sided_corr(self, frame: np.array, reference: np.array, limit):
        lag_pos = self._corr(frame, reference, limit // 2 - 1)
        lag_neg = self._corr(reference, frame, limit // 2)
        return lag_pos if lag_pos >= lag_neg else -lag_neg

    @staticmethod
    def _corr(frame: np.array, reference: np.array, end: int, start: int = 0) -> np.array:
        return start + np.argmax([reference[:len(reference) - i].dot(frame[i:]) for i in range(start, end)])

    @staticmethod
    def choose_idx(frames: pd.Series):
        variances = frames.apply(np.var)
        # mid = variances.mean()
        mid = variances.min()
        return np.abs(variances - mid).idxmin()



import numpy as np
import pandas as pd

from straw.io.params import StreamParams


class BaseCorrector:
    def apply(self, samplebuffer: np.ndarray, params: StreamParams) -> (np.ndarray, np.ndarray):
        """
        Apply a correction
        :param samplebuffer: array to apply the corrections to - ndarray with dimensions (channels, samples)
        :param params: stream params where the correction params should be stored
        :return: None
        """
        return

    def apply_revert(self, samplebuffer: np.ndarray, params: StreamParams) -> (np.ndarray, np.ndarray):
        """
        Apply the reverse of a correction
        :param samplebuffer: array to apply the corrections to - ndarray with dimensions (channels, samples)
        :param params: stream params where the correction params are be stored
        :return: None
        """
        return

    def df_wrap_apply(self, frameset: pd.Series):
        """
        DataFrame wrapper
        Used mainly for testing and evaluating localized corrections
        :param frameset: Series of frames
        :return: None
        """
        ndarr = np.stack(frameset.tolist())
        self.apply(ndarr, StreamParams())
        for i, idx in enumerate(frameset.index):
            frameset[idx][:] = ndarr[i]



import pandas as pd

from straw.lpc import steps
from straw.static import SubframeType

"""
Pandas-lever wrappers
"""


def compute_qlp(df: pd.DataFrame, order: int, qlp_coeff_precision: int) -> pd.DataFrame:
    """
    Compute LPC and quantize the LPC coefficients
    :param df: input dataframe with columns [frame]
    :param order: maximal LPC order
    :param qlp_coeff_precision: Bit precision for storing the quantized LPC coefficients
    :return: Series(qlp coefficients, qlp precision, qlp shift)
    """
    # frame["qlp"] = None
    # frame["qlp"] = frame["qlp"].astype(object)
    # frame["qlp_precision"] = 0
    # frame["shift"] = 0
    #
    # for idx, row in frame.iterrows():
    #     lpc = steps.compute_lpc(row["frame"], p=order)
    #     qlp, precision, shift = steps.quantize_lpc_cython(lpc, qlp_coeff_precision)
    #     frame["qlp_precision"][idx] = precision
    #     frame["shift"][idx] = shift
    #     frame["qlp"][idx] = qlp
    #
    # return frame

    df["qlp"] = None
    df["qlp_precision"] = 0
    df["shift"] = 0

    # Constant frames
    const_frames = df["frame"].apply(lambda x: not (x - x[0]).any())
    if const_frames.any():
        df.loc[const_frames, "frame_type"] = SubframeType.CONSTANT
        df.loc[~const_frames, "frame_type"] = SubframeType.RAW
        return df

    lpc = steps.compute_lpc(df["frame"], order)
    if lpc is None or not steps.lpc_is_stable(lpc):
        df["frame_type"] = SubframeType.RAW
        return df

    qlp, precision, shift = steps.quantize_lpc_cython(lpc, qlp_coeff_precision)
    df["frame_type"] = SubframeType.LPC_COMMON
    df["qlp"] = [qlp for _ in range(len(df["qlp"]))]
    df["qlp_precision"] = precision
    df["shift"] = shift
    return df


def compute_residual(df: pd.DataFrame):
    """
    Computes the residual from the given signal with quantized LPC coefficients
    :param df: input dfframe slice with columns [frame, qlp, shift]
    :return: the input dfframe slice with a [residual] column added
    """

    def wrap(dt: pd.DataFrame):
        if not (dt["frame_type"] in (SubframeType.LPC, SubframeType.LPC_COMMON)):
            return None
        return steps.predict_compute_residual(dt["frame"], dt["qlp"], dt["shift"])

    df["residual"] = df.apply(wrap, axis=1)
    df.loc[df["residual"].isna() & df["frame_type"].isin(
        (SubframeType.LPC, SubframeType.LPC_COMMON)), "frame_type"] = SubframeType.RAW


def compute_original(df: (pd.Series, pd.DataFrame), inplace=False):
    """
    Computes the original from the given residual signal with quantized LPC coefficients and warmup samples
    :param df: input dfframe with columns [frame, qlp, shift]
    :param inplace: whether the restoring should be done in place (faster)
    this ignores the residual column and expects the residual to be stored after the warmup samples in the frame column
    :return: residual as a numpy array
    """
    if isinstance(df, pd.DataFrame):
        return df.apply(compute_original, inplace=inplace, axis=1)

    if not (df["frame_type"] in (SubframeType.LPC, SubframeType.LPC_COMMON)):
        return df

    if inplace:
        frame = df["frame"]
    else:
        frame = df["frame"].copy()
        frame[len(df["qlp"]):] = df["residual"]
        df["restored"] = frame

    steps.restore_signal_cython(frame, df["qlp"], df["shift"])

    return df



from functools import partial
from multiprocessing import Pool, cpu_count

import numpy as np
import pandas as pd
from pandas.core.groupby import DataFrameGroupBy


class ParallelCompute:
    """
    Class used to access the thread pool
    This class is a singleton
    """
    __instance = None

    cpus: int
    _apply_args: tuple
    _apply_kwargs: dict

    @staticmethod
    def get_instance():
        """
        Get the instance of this class
        :return:
        """
        if ParallelCompute.__instance is None:
            ParallelCompute()
        return ParallelCompute.__instance

    def __init__(self, cpus=cpu_count()):
        """
        Private constructor
        :param cpus: number of CPUs to use when parallel is True, None means use all
        """
        if ParallelCompute.__instance is not None:
            raise RuntimeError("This class is a singleton!")
        else:
            ParallelCompute.__instance = self
        self.cpus = cpus

    def _parallelize(self, data, func):
        with Pool(self.cpus) as pool:
            data = pool.map(func, np.array_split(data, self.cpus))
        return pd.concat(data)

    def _run_on_subset(self, func, data_subset):
        return data_subset.apply(func, args=self._apply_args, **self._apply_kwargs)

    def map(self, data, func: callable, args=(), **kwargs) -> pd.Series:
        """
        Apply the function to the given DataFrame or Series in parallel
        :param data: DataFrame or Series to which func will be applied
        :param func: function to apply
        :param args: args to use in apply
        :param kwargs: kwargs to use in apply
        :return: DataFrame or Series with applied data
        """
        self._apply_args = args
        self._apply_kwargs = kwargs
        return self._parallelize(data, partial(self._run_on_subset, func))

    def _group_parallelize(self, data, func):
        chunksize = max(int(len(data) / self.cpus) * 2, 1)
        with Pool(self.cpus + 1) as p:
            ret_list = p.map(func, [group for name, group in data], chunksize=chunksize)
        if not ret_list:
            return None
        elif isinstance(ret_list[0], (pd.Series, pd.DataFrame)):
            return pd.concat(ret_list)
        else:
            return ret_list

    def _group_run_on_subset(self, func, data_subset):
        return func(data_subset, *self._apply_args, **self._apply_kwargs)

    def map_group(self, data: DataFrameGroupBy, func: callable, args=(), **kwargs) -> pd.DataFrame:
        """
        Apply the function to the given DataFrame groupby object in parallel
        :param data: DataFrame to which func will be applied
        :param func: function to apply
        :param args: args to use in apply
        :param kwargs: kwargs to use in apply
        :return: DataFrame or Series with applied data
        """
        self._apply_args = args
        self._apply_kwargs = kwargs
        return self._group_parallelize(data, partial(self._group_run_on_subset, func))

    def _ndarray_parallelize(self, data, func):
        with Pool(self.cpus) as p:
            ret_list = p.map(func, data)
        return ret_list

    def map_ndarray(self, data: np.array, func: callable, args=(), **kwargs):
        """
        Apply the functiom to the given numpy array in parallel
        :param data: numpy array to which func will be applied
        :param func: function to apply
        :param args: args to use in apply
        :param kwargs: kwargs to use in apply
        :return: DataFrame or Series with applied data
        """
        self._apply_args = args
        self._apply_kwargs = kwargs
        return self._ndarray_parallelize(data, partial(self._group_run_on_subset, func))



import numpy as np
import pandas as pd
from bitarray import bitarray
from bitarray.util import int2ba
from tqdm import tqdm

from straw.io.base import BaseWriter, BaseReader
from straw.io.sizes import StrawSizes
from straw.static import SubframeType
from . import ext_io

"""
This file contains the Straw format witing/reading interface
The format is witten using recursive descent
"""


class StrawFormatWriter(BaseWriter):
    def _format_specific_checks(self):
        if self._params.channels < 1:
            raise ValueError("No channels present!")

    def _stream(self):
        self._f.write("sTrW".encode("utf-8"))
        self._metadata_block()

    def _metadata_block(self):
        sec = self._metadata_block_header()
        sec += self._metadata_block_streaminfo()
        sec.tofile(self._f)

    def _metadata_block_header(self) -> bitarray:
        sizes = StrawSizes.metadata_block_header
        sec = bitarray()
        sec.append(1)  # this block is the last metadata block before the audio blocks
        sec += int2ba(0, length=sizes.type)  # BLOCK_TYPE: STREAMINFO
        # ignore if streaminfo...
        # sec += int2ba(data_size, length=24)
        return sec

    def _metadata_block_streaminfo(self) -> bitarray:
        sizes = StrawSizes.metadata_block_streaminfo
        sec = bitarray()
        sec += int2ba(self._params.sample_rate, length=sizes.samplerate)
        sec.frombytes(self.encode_int_utf8(self._params.channels - 1))
        sec += int2ba(self._params.bits_per_sample - 1, length=sizes.bps)
        sec += int2ba(self._params.total_frames, length=sizes.frames)
        sec += int2ba(self._params.total_samples, length=sizes.samples)
        sec.frombytes(self._params.md5)
        sec += int2ba(self._params.responsiveness, length=sizes.responsiveness)

        # Shift
        if self._params.lags.any():
            sec.append(1)
            sec.frombytes(self.encode_int_utf8(self._params.leading_channel))
            for val in self._params.lags:
                sec += int2ba(int(val), length=sizes.shift)
            for c in range(self._params.channels):
                for val in self._params.removed_samples_start[c]:
                    sec += int2ba(int(val), length=self._params.bits_per_sample, signed=True)
                for val in self._params.removed_samples_end[c]:
                    sec += int2ba(int(val), length=self._params.bits_per_sample, signed=True)
        else:
            sec.append(0)

        # Bias
        if self._params.bias.any():
            sec.append(1)
            for val in self._params.bias:
                sec += int2ba(int(val), length=sizes.bias, signed=True)
        else:
            sec.append(0)

        # Gain
        if self._params.gain.any():
            sec.append(1)
            for val in self._params.gain:
                sec += int2ba(int(val), length=sizes.gain)
            sec += int2ba(self._params.gain_shift, length=sizes.gain_shift)
        else:
            sec.append(0)

        sec.fill()
        return sec

    def _frame(self, df: pd.DataFrame):
        footer_sizes = StrawSizes.frame_footer
        sec = bitarray()
        tmp = df.apply(self._subframe, axis=1)
        for subframe_bitstream in tmp:
            sec += subframe_bitstream
        sec.fill()  # zero-padding to byte alignment

        # Header
        sec = self._frame_header(df, len(sec.tobytes()) + footer_sizes.crc // 8) + sec

        # Footer
        sec += int2ba(self.Crc.crc16(sec.tobytes()), length=footer_sizes.crc)
        return sec

    def _frame_header(self, df: pd.DataFrame, frame_data_size: int) -> bitarray:
        sizes = StrawSizes.frame_header
        sec = bitarray()
        sec += int2ba(0b10101010101010, length=sizes.sync_code)  # sync code
        sec.append(0)

        # block size start
        blocksize = int(df["frame"].apply(len).max())
        if blocksize == 0:
            raise ValueError("Blocksize cannot be 0")
        tmp = np.log2(blocksize)
        if tmp % 1 == 0.0:
            sec.append(0)  # get 8 bit exponent for (2^n) samples
            sec += int2ba(int(tmp), length=sizes.block_size_log2)
        else:
            sec.append(1)  # get 16 bit (blocksize-1)
            sec += int2ba(blocksize - 1, length=sizes.block_size_exact)
        # block size end

        sec.frombytes(self.encode_int_utf8(df.index.min()))
        sec += int2ba(frame_data_size + len(sec.tobytes()) + sizes.frame_bytes // 8 + sizes.crc // 8,
                      length=sizes.frame_bytes)
        sec += int2ba(self.Crc.crc8(sec.tobytes()), length=sizes.crc)
        return sec

    def _subframe(self, df: pd.Series) -> bitarray:
        sec = self._subframe_header(df)
        sec += self._subframe_data(df)
        return sec

    def _subframe_header(self, df: pd.Series) -> bitarray:
        sizes = StrawSizes.subframe_header
        sec = bitarray()
        sec += int2ba(df["frame_type"], length=sizes.type)
        return sec

    def _subframe_data(self, df: pd.Series) -> bitarray:
        subframe_type = df["frame_type"]
        if subframe_type == SubframeType.CONSTANT:  # SUBFRAME_CONSTANT
            return self._subframe_constant(df)
        elif subframe_type == SubframeType.RAW:  # SUBFRAME_RAW
            return self._subframe_raw(df)
        elif subframe_type == SubframeType.LPC:  # SUBFRAME_LPC
            return self._subframe_lpc(df)
        elif subframe_type == SubframeType.LPC_COMMON:  # SUBFRAME_LPC_COMMON
            if df["channel"] == 0:
                return self._subframe_lpc(df)
            else:
                return self._subframe_lpc_common(df)
        else:
            raise ValueError(f"Invalid frame type: {subframe_type}")

    def _subframe_constant(self, df: pd.Series) -> bitarray:
        sec = bitarray()
        sec += int2ba(int(df["frame"][0]), length=self._params.bits_per_sample, signed=True)
        return sec

    def _subframe_raw(self, df: pd.Series) -> bitarray:
        buffer = np.zeros(df["frame"].shape[0] * (self._params.bits_per_sample // 8), dtype=np.uint8)
        ext_io.write_frame(buffer, df["frame"] + self._params.bias[df["channel"]], self._params.bits_per_sample)
        sec = bitarray(buffer=buffer)
        return sec

    def _subframe_lpc(self, df: pd.Series) -> bitarray:
        sizes = StrawSizes.subframe_lpc
        sec = bitarray()

        qlp = df["qlp"]
        order = len(qlp)
        qlp_precision = int(df["qlp_precision"])
        shift = int(df["shift"])

        sec += int2ba(order - 1, length=sizes.lpc_order)
        sec += int2ba(qlp_precision - 1, length=sizes.lpc_prec)
        sec += int2ba(shift, length=sizes.lpc_shift)
        for coeff in qlp:
            sec += int2ba(int(coeff), length=qlp_precision, signed=True)

        sec += self._subframe_lpc_common(df)
        return sec

    def _subframe_lpc_common(self, df: pd.Series):
        sec = bitarray()
        warmup_samples = df["frame"][:len(df["qlp"])] + self._params.bias[df["channel"]]
        for warmup_sample in warmup_samples:
            sec += int2ba(int(warmup_sample), length=self._params.bits_per_sample, signed=True)
        sec += self._residual(df)
        return sec

    def _residual(self, df: pd.Series) -> bitarray:
        sizes = StrawSizes.residual
        sec = bitarray()
        sec += int2ba(int(df["bps"]), length=sizes.param)
        sec += df["stream"]
        return sec


class StrawFormatReader(BaseReader):
    def _stream(self):
        marker = self._sec.get_bytes(32)
        if marker.decode("utf-8") != "sTrW":
            raise ValueError("Not a valid Straw file!")
        expected_frames = self._metadata_block()
        if self.show_progress:
            pbar = tqdm(range(expected_frames))
            pbar.set_description(f"Loading frames")
        else:
            pbar = range(expected_frames)
        for _ in pbar:
            if self._sec.is_eof():
                pbar.close()
                break

            self._frame(expected_frames)

    def _metadata_block(self) -> int:
        data_len = self._metadata_block_header()
        return self._metadata_block_streaminfo()

    def _metadata_block_header(self) -> int:
        sizes = StrawSizes.metadata_block_header
        last_metadata_block = self._sec.get_int()  # this block is the last metadata block before the audio blocks
        block_type = self._sec.get_int(length=sizes.type)  # BLOCK_TYPE: STREAMINFO
        if block_type == 0:
            return 0
        else:
            data_len = self._sec.get_int(length=sizes.size)
            return data_len

    def _metadata_block_streaminfo(self) -> int:
        sizes = StrawSizes.metadata_block_streaminfo
        self._params.sample_rate = self._sec.get_int(length=sizes.samplerate)
        self._params.channels = self._sec.get_int_utf8() + 1
        self._params.alloc_arrays()
        self._params.bits_per_sample = self._sec.get_int(length=sizes.bps) + 1
        expected_frames = self._sec.get_int(length=sizes.frames)
        self._params.total_samples = self._sec.get_int(length=sizes.samples)
        self._params.md5 = self._sec.get_bytes(length=sizes.md5)
        self._params.responsiveness = self._sec.get_int(length=sizes.responsiveness)
        self._ricer.responsiveness = self._params.responsiveness

        # Allocate sample buffer
        self._allocate_buffer()

        # Shift
        has_shift = self._sec.get_int()
        if has_shift:
            self._params.leading_channel = self._sec.get_int_utf8()
            for i in range(self._params.lags.shape[0]):
                self._params.lags[i] = self._sec.get_int(length=sizes.shift)
            total_size = self._samplebuffer.shape[1] - np.max(self._params.lags)
            for c in range(self._params.channels):
                lag = self._params.lags[c]
                for s in range(lag):
                    self._samplebuffer[c][s] = self._sec.get_int(length=self._params.bits_per_sample, signed=True)
                for e in range(total_size + lag, self._samplebuffer.shape[1]):
                    self._samplebuffer[c][e] = self._sec.get_int(length=self._params.bits_per_sample, signed=True)

        # Bias
        has_bias = self._sec.get_int()
        if has_bias:
            for i in range(self._params.bias.shape[0]):
                self._params.bias[i] = self._sec.get_int(length=sizes.bias, signed=True)

        # Gain
        has_gain = self._sec.get_int()
        if has_gain:
            for i in range(self._params.gain.shape[0]):
                self._params.gain[i] = self._sec.get_int(length=sizes.gain)
            self._params.gain_shift = self._sec.get_int(length=sizes.gain_shift)

        self._sec.skip_padding()
        return expected_frames

    def _frame(self, expected_frames: int):
        start = self._sec.get_pos()
        seq, frame_size, blocksize = self._frame_header()
        last_lpc_frame = 0
        for i in range(self._params.channels):
            row = self._subframe(last_lpc_frame, blocksize, i)
            row["seq"] = seq
            row["idx"] = i * expected_frames + seq
            # we expect that the first channel will have lpc coefficients in every case
            if "qlp" in row and isinstance(row["qlp"], np.ndarray):
                last_lpc_frame = row
            self._raw.append(row)
        self._samplebuffer_ptr += blocksize
        self._sec.skip_padding()

        # Footer
        footer_sizes = StrawSizes.frame_footer
        expected_crc = self.Crc.crc16(self._sec.get_from(start).tobytes())
        checksum = self._sec.get_int(length=footer_sizes.crc)
        if expected_crc != checksum:
            raise RuntimeError(f"Inavalid frame checksum at frame {seq}")

    def _frame_header(self) -> (dict, int, int):
        sizes = StrawSizes.frame_header
        start = self._sec.get_pos()
        if self._sec.get_int(length=sizes.sync_code) != 0b10101010101010:  # sync code
            raise RuntimeError("Lost sync")
        self._sec.get_int()
        if self._sec.get_int():
            blocksize = self._sec.get_int(length=sizes.block_size_exact) + 1
        else:
            blocksize = 1 << self._sec.get_int(length=sizes.block_size_log2)

        seq = self._sec.get_int_utf8()
        frame_size = self._sec.get_int(length=sizes.frame_bytes)

        expected_crc = self.Crc.crc8(self._sec.get_from(start).tobytes())
        checksum = self._sec.get_int(length=sizes.crc)
        if expected_crc != checksum:
            raise RuntimeError(f"Inavalid frame header checksum at frame {seq}")
        return seq, frame_size, blocksize

    def _subframe(self, last_lpc: dict, blocksize: int, subframe_num: int) -> dict:
        subframe_type = self._subframe_header()
        return self._subframe_data(subframe_type, last_lpc, blocksize, subframe_num)

    def _subframe_header(self) -> int:
        sizes = StrawSizes.subframe_header
        return self._sec.get_int(length=sizes.type)

    def _subframe_data(self, subframe_type: int, last_lpc: dict, blocksize: int, subframe_num: int) -> dict:
        if subframe_type == SubframeType.CONSTANT:  # SUBFRAME_CONSTANT
            return self._subframe_constant(blocksize, subframe_num)
        elif subframe_type == SubframeType.RAW:  # SUBFRAME_RAW
            return self._subframe_raw(blocksize, subframe_num)
        elif subframe_type == SubframeType.LPC:  # SUBFRAME_LPC
            return self._subframe_lpc(blocksize, subframe_num)
        elif subframe_type == SubframeType.LPC_COMMON:  # SUBFRAME_LPC
            if subframe_num == 0:
                return self._subframe_lpc(blocksize, subframe_num)
            else:
                return self._subframe_lpc_common(last_lpc, blocksize, subframe_num)
        else:
            raise ValueError(f"Invalid frame type: {subframe_type}")

    def _subframe_constant(self, blocksize: int, subframe_num: int):
        row = {
            "frame_type": SubframeType.CONSTANT,
            "channel": subframe_num,
            "frame": self._samplebuffer[subframe_num][
                     self._samplebuffer_ptr + self._params.lags[subframe_num]:self._samplebuffer_ptr +
                                                                              self._params.lags[
                                                                                  subframe_num] + blocksize]
        }
        row["frame"][:] = self._sec.get_int(length=self._params.bits_per_sample, signed=True)
        return row

    def _subframe_raw(self, blocksize: int, subframe_num: int) -> dict:
        row = {
            "frame_type": SubframeType.RAW,
            "channel": subframe_num,
            "frame": self._samplebuffer[subframe_num][
                     self._samplebuffer_ptr + self._params.lags[subframe_num]:self._samplebuffer_ptr +
                                                                              self._params.lags[
                                                                                  subframe_num] + blocksize]
        }
        bitbytes = self._sec.get_bytes(length=self._params.bits_per_sample * blocksize)
        ext_io.read_frame(row["frame"], bitbytes, self._params.bits_per_sample, 1)
        row["frame"] -= self._params.bias[subframe_num]
        return row

    def _subframe_lpc(self, blocksize: int, subframe_num: int) -> dict:
        sizes = StrawSizes.subframe_lpc
        row = {
            "frame_type": SubframeType.LPC,
            "channel": subframe_num,
            "frame": self._samplebuffer[subframe_num][
                     self._samplebuffer_ptr + self._params.lags[subframe_num]:self._samplebuffer_ptr +
                                                                              self._params.lags[
                                                                                  subframe_num] + blocksize],
            "qlp_precision": 0,
            "shift": 0,
            "qlp": None,
        }

        order = self._sec.get_int(length=sizes.lpc_order) + 1
        row["qlp_precision"] = self._sec.get_int(length=sizes.lpc_prec) + 1
        row["shift"] = self._sec.get_int(length=sizes.lpc_shift)
        row["qlp"] = np.asarray(
            [self._sec.get_int(length=row["qlp_precision"], signed=True) for _ in range(order)], dtype=np.int32)

        for i in range(order):
            warmup_sample = self._sec.get_int(length=self._params.bits_per_sample, signed=True)
            row["frame"][i] = warmup_sample - self._params.bias[subframe_num]

        # get the residual
        row["residual"] = self._samplebuffer[subframe_num][
                          self._samplebuffer_ptr + self._params.lags[subframe_num] + order:self._samplebuffer_ptr +
                                                                                           self._params.lags[
                                                                                               subframe_num] + blocksize]
        row["residual"], row["bps"] = self._residual(array=row["residual"])
        return row

    def _subframe_lpc_common(self, last_lpc: dict, blocksize: int, subframe_num: int) -> dict:
        row = {
            "frame_type": SubframeType.LPC_COMMON,
            "channel": subframe_num,
            "frame": self._samplebuffer[subframe_num][
                     self._samplebuffer_ptr + self._params.lags[subframe_num]:self._samplebuffer_ptr +
                                                                              self._params.lags[
                                                                                  subframe_num] + blocksize],
            "qlp_precision": last_lpc["qlp_precision"],
            "shift": last_lpc["shift"],
            "qlp": last_lpc["qlp"],
        }

        last_lpc["frame_type"] = SubframeType.LPC_COMMON
        order = len(last_lpc["qlp"])
        for i in range(order):
            warmup_sample = self._sec.get_int(length=self._params.bits_per_sample, signed=True)
            row["frame"][i] = warmup_sample - self._params.bias[subframe_num]

        # get the residual
        row["residual"] = self._samplebuffer[subframe_num][
                          self._samplebuffer_ptr + self._params.lags[subframe_num] + order:self._samplebuffer_ptr +
                                                                                           self._params.lags[
                                                                                               subframe_num] + blocksize]
        row["residual"], row["bps"] = self._residual(array=row["residual"])
        return row

    def _residual(self, array: np.array) -> np.array:
        sizes = StrawSizes.residual
        bps = self._sec.get_int(length=sizes.param)
        bits_read = self._ricer.bitstream_to_frame(
            self._memview,
            len(array), bps, own_frame=array, bitarray_pos=self._sec.get_pos())
        self._sec.advance(bits_read)
        return array, bps



from hashlib import md5

import numpy as np
import pandas as pd

from straw.io.params import StreamParams


class BaseCoder:
    """
    Base class for the Encoder and Decoder
    """
    # Member utils
    _flac_mode: bool

    # Member variables
    _default_frame_size: int = 1 << 12  # bytes
    _data: pd.DataFrame
    _params: StreamParams
    _samplebuffer: np.ndarray

    _source_size: int

    _supported_subtypes = {
        "PCM_16": 16,
        # "PCM_24": 24,
        # "PCM_32": 32,
    }

    _subtype_pattern = "PCM_{}"

    def __init__(self, flac_mode=False, show_progress: bool = False):
        """
        Common constructor for Encoder and Decoder
        :param flac_mode: False - FLAC mode is not supported anymore
        :param show_progress: show encoding progress - currently has no effect for the encoder
        """
        if flac_mode:
            raise NotImplementedError("FLAC mode is not supported anymore")
        self._flac_mode = flac_mode
        self.show_progress = show_progress

    def _slice_channel_data_into_frames(self, data: np.array, limits: np.array = None):
        """
        Slice the given 1D data into a list of numpy views
        :param data: input samplebuffer
        :param limits: optional slicing points, if None then the default block size is used
        :return: list of numpy arrays (views)
        """
        if limits is None:
            return [data[i:i + self._default_frame_size] for i in range(0, len(data), self._default_frame_size)]
        else:
            return [data[limits[i]:limits[i + 1]] for i in range(limits.shape[0] - 1)]

    def usage_mib(self):
        """
        Returns the deep memory usage of the given dataframe in mebibytes
        :return: memory usage of the given dataframe
        """
        return self._data.memory_usage(index=True, deep=True).sum() / (2 ** 20)

    def sample_frame(self, seq=0) -> pd.Series:
        """
        Get a sample frame - one channel - series
        :param seq: the sequence number of the requested frame
        :return: sample frame
        """
        return self._data.loc[seq]

    def sample_frame_multichannel(self, seq=0) -> pd.DataFrame:
        """
        Get a sample frame - multiple channels - dataframe
        :param seq: the sequence number of the requested frame
        :return: sample frame
        """
        return self._data[self._data["seq"] == seq]

    def samplebuffer_frame_multichannel(self, seq=0, blocksize=4096) -> np.array:
        """
        Get a sample frame - multiple channels - array
        :param seq: the sequence number of the requested frame
        :param blocksize: presumed size of blocks when slicing
        :return: sample frame
        """
        return self._samplebuffer[:, seq * blocksize:seq * blocksize + blocksize].copy()

    def get_data(self) -> pd.DataFrame:
        """
        Getter for data
        :return: data
        """
        return self._data

    def get_soundfile_compatible_array(self) -> np.array:
        """
        Get a soundfile-compatible numpy view of the samplebuffer
        :return: soundfile-compatible numpy array
        """
        return self._samplebuffer.swapaxes(1, 0)

    def get_params(self) -> StreamParams:
        """
        Getter for params
        :return: params
        """
        return self._params

    def get_md5(self) -> bytes:
        """
        Calculates the md5sum of the current samplebuffer
        :return: md5 digests
        """
        try:
            return md5(self._samplebuffer.swapaxes(1, 0)).digest()
        except ValueError:
            return md5(np.ascontiguousarray(self._samplebuffer.swapaxes(1, 0))).digest()



import unittest

import numpy as np

from straw import rice
from . import resources


class Rice(unittest.TestCase):
    """
    Rice parameter 4
    """
    k = 2
    m = 1 << k
    r = rice.Ricer()

    def test_rice_ext_arange_decode(self):
        frame = np.arange(20, dtype=np.int16)

        bits = self.r.frame_to_bitstream(frame, self.k)
        decoded = self.r.bitstream_to_frame(memoryview(bits).tobytes(), len(frame), self.k)

        self.assertListEqual(frame.tolist(), decoded.tolist())

    def test_rice_ext_signal_encode_smallbuffer(self):
        # This frame would produce a residual larger than the given buffer, so it would be encoded raw
        frame = resources.get_signal()[0]

        max_allowed_bits = frame.nbytes * 8
        bits = self.r.frame_to_bitstream(frame, self.k)

        self.assertEqual(len(bits), max_allowed_bits)

    def test_problematic_residual1(self):
        frame = resources.get_problematic_residual1()

        bits = self.r.frame_to_bitstream(frame, self.k)
        decoded = self.r.bitstream_to_frame(memoryview(bits).tobytes(), len(frame), self.k)

        self.assertListEqual(frame.tolist(), decoded.tolist())

    def test_problematic_residual2(self):
        frame = resources.get_problematic_residual2()

        bits = self.r.frame_to_bitstream(frame, self.k)
        decoded = self.r.bitstream_to_frame(memoryview(bits).tobytes(), len(frame), self.k)

        self.assertListEqual(frame.tolist(), decoded.tolist())

if __name__ == '__main__':
    unittest.main()



import unittest

import numpy as np
import pandas as pd

from straw import lpc
from straw.static import SubframeType
from . import resources


class LPCSignalIntegrity(unittest.TestCase):
    lpc_order = 8
    lpc_precision = 12  # bits

    signal, sr = resources.get_signal()
    signal_zeros = np.zeros(2 ** 12)

    def test_reconstruction(self):
        df = pd.DataFrame({"frame": [self.signal.copy()], "frame_type": SubframeType.LPC})
        lpc.compute_qlp(df, self.lpc_order, self.lpc_precision)

        lpc.compute_residual(df)
        df = lpc.compute_original(df, inplace=False)

        self.assertEqual((self.signal - df.loc[0, "restored"]).any(), False)

    def test_zeroframe_qlp(self):
        df = pd.DataFrame({"frame": [self.signal_zeros], "frame_type": SubframeType.LPC})
        lpc.compute_qlp(df, self.lpc_order, self.lpc_precision)
        self.assertEqual(df.loc[0, "shift"], 0)
        self.assertEqual(df.loc[0, "qlp_precision"], 0)

    def test_zeroframe_residual(self):
        df = pd.DataFrame({"frame": [self.signal_zeros], "frame_type": SubframeType.LPC})
        lpc.compute_qlp(df, self.lpc_order, self.lpc_precision)

        lpc.compute_residual(df)
        self.assertIsNone(df.loc[0, "qlp"])
        self.assertIsNone(df.loc[0, "residual"])
        self.assertEqual(df.loc[0, "frame_type"], SubframeType.CONSTANT)


if __name__ == '__main__':
    unittest.main()



import numpy as np

"""
Parameters of the Straw and FLAC format
"""


class StreamParams:
    def alloc_arrays(self):
        if self.channels is None:
            raise ValueError("Channels must be initialized when allocate_arrays is called")
        self.lags = np.zeros(self.channels, dtype=np.int8)  # 4-bit
        self.bias = np.zeros(self.channels, dtype=np.int8)  # 8-bit signed
        self.gain = np.zeros(self.channels, dtype=np.int64)  # quantized floating point number with variable bit width

    sample_rate: int = None
    channels: int = None
    bits_per_sample: int = None
    total_samples: int = None
    total_frames: int = None
    md5: bytes = None
    responsiveness: int = None
    leading_channel: int = None
    lags: np.ndarray  # length equal to number of channels, each value represents lag on a channel
    removed_samples_start = []  # removed samples for each channel from the start
    removed_samples_end = []  # removed samples for each channel from the end
    bias: np.ndarray  # length equal to number of channels, each value represents bias on a channel
    gain: np.ndarray  # length equal to number of channels, each value represents quantized gain factor on a channel
    gain_shift: int = None


class FLACStreamParams(StreamParams):
    min_block_size: int = None
    max_block_size: int = None
    min_frame_size: int = None
    max_frame_size: int = None



import numpy as np
import pandas as pd
from bitarray import bitarray

from straw.compute import ParallelCompute
from straw.io.sizes import StrawSizes
from straw.static import SubframeType, Default
from . import ext_rice


class Ricer:
    """
    Rice encoder/decoder
    Currently only supports memory for for memory efficiency comparisons and benchmarks
    """

    def __init__(self, adaptive=True, responsiveness: int = Default.framing_resolution):
        self.adaptive = adaptive
        self.parallel = ParallelCompute.get_instance()
        self.responsiveness = responsiveness

    ##########################
    # Optimal order guessing #
    ##########################

    def guess_parameter(self, frame_residual: np.array) -> np.int8:
        """
        Estimate the best starting encoding parameter
        :param frame_residual: residual to be encoded
        :return: starting k parameter
        """
        if frame_residual is None:
            return np.int8(0)
        smallframe = frame_residual[:self.responsiveness].astype(np.int64)
        ext_rice.interleave_frame(smallframe)
        param = np.clip(np.log2(smallframe.mean()), 0, (1 << StrawSizes.residual.param) - 1)
        return np.round(param).astype(np.int8)

    ############
    # Encoding #
    ############

    def frame_to_bitstream(self, frame: np.array, bps: int) -> bitarray:
        """
        Encode a numpy frame to a bitsream
        :param frame: numpy array of samples to encode
        :param bps: expected bits per sample
        :return: encoded bitarray
        """
        data = np.zeros(frame.nbytes, dtype=np.uint8)
        bits = ext_rice.encode_frame(data, frame, bps, self.responsiveness, adaptive=self.adaptive)
        if bits == -1:
            return bitarray(buffer=frame)  # This will later fail in Encoder._ensure_compression
        else:
            return bitarray(buffer=data)[:bits]

    def _frame_to_bitstream_df_expander(self, df: pd.DataFrame) -> np.array:
        if df["frame_type"] not in (SubframeType.LPC, SubframeType.LPC_COMMON):
            return bitarray()

        return self.frame_to_bitstream(df["residual"], df["bps"])

    def frames_to_bitstreams(self, df: pd.DataFrame, parallel: bool = True) -> pd.Series:
        """
        Encode frames to a series of bitsreams
        :param df: DataFrame with columns ["residual", "bps", "frame_type"]
        :param parallel: if True then use multithreading
        :return: encoded bitarrays
        """

        if not parallel:
            return df.apply(self._frame_to_bitstream_df_expander, axis=1, result_type="reduce")

        return self.parallel.map(df, self._frame_to_bitstream_df_expander, axis=1, result_type="reduce")

    ############
    # Decoding #
    ############

    def bitstream_to_frame(self, bitstream_memoryview: bytes,
                           frame_size: int, bps: int,
                           own_frame: np.array = None,
                           bitarray_pos=0):
        """
        Decode a single frame from a given bitstream
        WARNING: The given bitstream is destroyed to prevent unnecessary memory duplication
        :param bitstream_memoryview: rice encoded stream
        :param frame_size: frame size
        :param bps: expected bits per sample
        :param own_frame: if not None, this frame will be filled
        :param bitarray_pos: starting position within the bitarray - use this to avoid slicing
        :return: decoded frame or number of bits read if own_frame is not None
        """
        bits_read = 0
        if own_frame is None:
            frame = np.zeros(frame_size, dtype=np.int64)
        else:
            frame = own_frame[:frame_size]

        if len(bitstream_memoryview) > 0:
            bits_read = ext_rice.decode_frame(frame, bitstream_memoryview, bps, self.responsiveness,
                                              adaptive=self.adaptive, starting_i=bitarray_pos)

        if own_frame is None:
            return frame
        else:
            return bits_read

    def _bitstream_to_frame_df_expander(self, df: pd.DataFrame) -> np.array:
        return self.bitstream_to_frame(df["stream"], df["size"], df["bps"])

    def bitstreams_to_frames(self, bitstreams: pd.Series, frame_sizes: pd.Series, bps: pd.Series,
                             parallel: bool = True) -> pd.Series:
        """
        Encode a series of bitstreams to a series of frames
        :param bitstreams: series of bitstreams
        :param frame_sizes: series of frame sizes
        :param bps: expected bits per second for this frame
        :param parallel: if True then use multithreading
        :return: series of decoded frames
        """

        comp = pd.DataFrame({"stream": bitstreams, "size": frame_sizes, "bps": bps})

        if not parallel:
            return comp.apply(self._bitstream_to_frame_df_expander, axis=1, result_type="reduce")

        return self.parallel.map(comp, self._bitstream_to_frame_df_expander, axis=1, result_type="reduce")

    ###########
    # Utility #
    ###########

    @staticmethod
    def frame_to_kparams(frame: np.ndarray, k: int, responsiveness: int = 16):
        """
        Transforms a frame to the used k params
        :param frame: frame to be encoded
        :param k: the starting k parameter
        :param responsiveness: responsiveness value
        :return: k parameters in the frame
        """
        frame = np.copy(frame)
        ext_rice.kparams(frame, k, responsiveness)
        return frame

    @staticmethod
    def frame_to_interleaved(frame: np.ndarray):
        """
        Transforms a frame to interleaved values
        :param frame: frame to be interleaved
        :return: frame with interelaved values
        """
        frame = np.copy(frame)
        ext_rice.interleave_frame(frame)
        return frame



UTF8 = None


class StrawSizes:
    """
    Sizes for the straw format
    Only for fields with at least one known dimension length
    None for unknown values
    All sizes are in bits unless specified otherwise
    """

    class metadata_block_header:  # noqa
        last = 1
        type = 7
        size = 24

    class metadata_block_streaminfo:  # noqa
        samplerate = 20
        channels = UTF8
        bps = 5
        frames = 27
        samples = 36
        md5 = 128
        responsiveness = 8
        leading_channel = UTF8
        shift = 4
        removed_samples = None
        bias = 8
        gain = 12
        gain_shift = 4

    class frame_header:  # noqa
        sync_code = 14
        contains_lpc = 1
        block_size_length = 1
        block_size_log2 = 8
        block_size_exact = 16
        frame_num = UTF8
        frame_bytes = 32
        crc = 8

    class frame_footer:  # noqa
        crc = 16

    class subframe_header:  # noqa
        type = 2

    class subframe_lpc:  # noqa
        contains_lpc = 1
        lpc_order = 5
        lpc_prec = 4
        lpc_shift = 4
        lpc_coeffs = None
        is_coded = 1

    class residual:  # noqa
        param = 4



import sys
from pathlib import Path
from typing import TextIO

import numpy as np
import pandas as pd
import soundfile

from straw import lpc, static, correctors
from straw.codec.base import BaseCoder
from straw.compute import ParallelCompute
from straw.io import Formatter
from straw.io.params import StreamParams
from straw.rice import Ricer
from straw.static import SubframeType, Default
from straw.util import Signals


class EncoderStats:
    frames: int = 0
    file_size: int = 0
    ratio: int = 0


class Encoder(BaseCoder):
    _lpc_order = 20  # can be sourced from len(df["qlp"]) once per group
    _lpc_precision = 12  # bits, stored in df["qlp_precision"] once per group
    _params = StreamParams()

    ##########
    # Public #
    ##########

    def __init__(self,
                 flac_mode=False,
                 do_corrections=("shift", "bias"),
                 dynamic_blocksize=False,
                 min_block_size=Default.min_frame_size,
                 max_block_size=Default.max_frame_size,
                 framing_treshold=Default.framing_treshold,
                 framing_resolution=Default.framing_resolution,
                 responsiveness=Default.rice_responsiveness,
                 parallelize=True,
                 show_progress: bool = False):
        """
        Encoder constructor
        :param do_corrections: tuple of correction to be performed
        :param dynamic_blocksize: whether to perform dynamic block slicing
        :param min_block_size: minimal block size
        :param max_block_size: maximal block size
        :param framing_treshold: framing treshold
        :param framing_resolution: framing resolution
        :param responsiveness: Rice coding responsiveness
        :param parallelize: if True use parallelization while encoding
        """
        super(Encoder, self).__init__(flac_mode, show_progress=show_progress)
        self._ricer = Ricer(adaptive=True if not flac_mode else False, responsiveness=responsiveness)
        self._params.responsiveness = responsiveness
        self._do_corrections = do_corrections
        self._do_dynamic_blocking = dynamic_blocksize
        self.min_block_size = min_block_size
        self.max_block_size = max_block_size
        self.framing_treshold = framing_treshold
        self.framing_resolution = framing_resolution
        self.parallelize = parallelize

    def set_rice_responsiveness(self, responsiveness):
        """
        Set a new Rice coding responsiveness
        :param responsiveness: new Rice responsiveness
        :return: None
        """
        self._ricer.responsiveness = responsiveness
        self._params.responsiveness = responsiveness

    def load_file(self, file):
        """
        Load the specified file into memory
        :param file: str or int or file-like object - anything that soundfile accepts
        :return: None
        """
        self._source_size = Path(file).stat().st_size
        with soundfile.SoundFile(file, "r") as wav:
            subtype = wav.subtype
            if subtype not in self._supported_subtypes:
                raise ValueError(f"Subtype '{subtype}' not supported, must be one of {self._supported_subtypes.keys()}")
            bits_per_sample = self._supported_subtypes[subtype]
            dtype_bits = static.soundfile_dtype[bits_per_sample]
            data = wav.read(dtype=f"int{dtype_bits}", always_2d=True)
            data >>= dtype_bits - bits_per_sample
            sr = wav.samplerate

        self.load_data(data, sr, bits_per_sample)

    def load_data(self, data: np.array, samplerate: int, bits_per_sample: int):
        """
        Load the given data into the internal DataFrame
        :param data: data to be loaded
        :param samplerate: samplerate of the given data
        :param bits_per_sample: bits per sample of the original data (can be lower than the bit width of data.dtype)
        :return: None
        """
        if len(data.shape) == 1:
            self._samplebuffer = data.reshape((1, -1))
        elif data.shape[0] > data.shape[1]:
            self._samplebuffer = data.swapaxes(1, 0)
        else:
            self._samplebuffer = data

        self._params.channels = int(self._samplebuffer.shape[0])
        self._params.total_samples = int(self._samplebuffer.shape[1])
        self._params.bits_per_sample = bits_per_sample
        self._params.alloc_arrays()

        # If file size not available, use raw sample size
        if not hasattr(self, "_source_size"):
            self._source_size = self._params.bits_per_sample * self._params.total_samples * self._params.channels

        self._params.sample_rate = samplerate
        self._params.md5 = self.get_md5()
        correctors.apply_corrections(self._samplebuffer, self._do_corrections, self._params)
        self._create_dataframe()

    def encode(self):
        """
        Encode the data in the internal DataFrame
        :return: None
        """
        # Initialize frame types
        self._init_frame_types()

        groups = self._data.groupby("seq")
        if self.parallelize and self._params.channels > 1:
            self._data = ParallelCompute.get_instance().map_group(groups, self._encode_frame)
        else:
            self._data = groups.apply(self._encode_frame)

    def _encode_frame(self, data_slice: pd.DataFrame):
        """
        Encode one frame - group of subframes
        :param data_slice: slice of a DataFrame containing a group of subframes
        :return: modified data_slice
        """
        # correctors.ShiftCorrector().df_wrap_apply(data_slice["frame"])
        lpc.compute_qlp(data_slice, order=self._lpc_order, qlp_coeff_precision=self._lpc_precision)
        lpc.compute_residual(data_slice)
        # correctors.GainCorrector().df_wrap_apply(data_slice["residual"])
        # correctors.ShiftCorrector().df_wrap_apply(data_slice["residual"])
        # correctors.BiasCorrector().df_wrap_apply(data_slice["residual"])
        data_slice = correctors.Decorrelator().midside_decorrelate(data_slice, "residual")
        data_slice["bps"] = data_slice["residual"].apply(self._ricer.guess_parameter)
        data_slice["stream"] = self._ricer.frames_to_bitstreams(data_slice, parallel=False)
        data_slice["stream_len"] = data_slice["stream"].apply(len)
        data_slice["frame_type"] = self._should_be_raw_maxbytes(data_slice)
        return data_slice

    def save_file(self, output_file):
        """
        Save the encoded signal
        :param output_file: target file
        :return: None
        """
        self._params.total_frames = int(len(self._data[self._data["channel"] == 0]))
        opened = False
        if not hasattr(output_file, "write"):
            output_file = open(output_file, "wb")
            opened = True

        Formatter().save(self._data, self._params, output_file, self._flac_mode)
        if opened:
            output_file.close()

    ###########
    # Private #
    ###########

    def _create_dataframe(self):
        """
        Create a dataframe from the raw signal, this includes slicing the signal into specific views
        NOTE: the underlying memory stays as a contiguous memory chunk
        :return: None
        """

        total_size = self._samplebuffer.shape[1] - np.max(self._params.lags)
        if self._do_dynamic_blocking:
            lag = self._params.lags[0]
            limits = Signals.get_frame_limits_by_energy(self._samplebuffer[0][lag:total_size + lag],
                                                        min_block_size=self.min_block_size,
                                                        max_block_size=self.max_block_size,
                                                        treshold=self.framing_treshold,
                                                        resolution=self.framing_resolution)
        else:
            limits = None

        ds = {"seq": [], "frame": [], "channel": []}
        for channel, channel_data in enumerate(self._samplebuffer):
            lag = self._params.lags[channel]
            sliced = self._slice_channel_data_into_frames(channel_data[lag:total_size + lag], limits=limits)
            ds["seq"] += [i for i in range(len(sliced))]
            ds["frame"] += sliced
            ds["channel"] += [channel for _ in range(len(sliced))]

        self._data = pd.DataFrame(ds, copy=False)

    def _init_frame_types(self):
        # all frames are LPC frames by default
        self._data["frame_type"] = np.full(len(self._data["frame"]), SubframeType.LPC, dtype="B")

    def _should_be_raw_maxbytes(self, df: pd.DataFrame):
        if not (df["frame_type"].isin((SubframeType.LPC, SubframeType.LPC_COMMON))).all():
            return df["frame_type"]

        max_allowed_bits = df.loc[df.index[0], "residual"].shape[0] * self._params.bits_per_sample
        if (df["stream_len"] >= max_allowed_bits).any():
            df["frame_type"] = SubframeType.RAW
        return df["frame_type"]

    @staticmethod
    def _decorrelate_signals(data_slice, col_name="residual"):
        # self._data = self._data.groupby("seq").apply(Decorrelator().localized_decorrelate, col_name=col_name)
        # self._data = ParallelCompute.get_instance().map_group(self._data.groupby("seq"),
        #                                                       Decorrelator().midside_decorrelate, col_name=col_name)
        data_slice = data_slice.groupby("seq").apply(correctors.Decorrelator().midside_decorrelate, col_name=col_name)
        data_slice["was_coded"] = 0

    ###########
    # Utility #
    ###########

    def get_stats(self, output_file: Path) -> EncoderStats:
        """
        Return the stats of the encoding
        :param output_file: the output file for comparison
        :return:
        """
        stats = EncoderStats()
        stats.file_size = output_file.stat().st_size
        stats.ratio = output_file.stat().st_size / self._source_size
        stats.frames = len(self._data.groupby('seq').groups)
        return stats

    def print_stats(self, output_file: Path, stream: TextIO = sys.stdout):
        """
        Verbose output - print a bunch of stuff...
        :param output_file: output file (only the size is needed)
        :param stream: stream where the output should be written
        :return: None
        """
        print(f"Number of frames: {len(self._data.groupby('seq').groups)}", file=stream)
        print(f"Number of subframes: {len(self._data)}", file=stream)
        print(f"Source size: {self._source_size} ({self._source_size / (2 ** 20):.2f} MiB)", file=stream)
        size = self._data["stream_len"].sum()
        print(f"md5: {self._params.md5.hex(' ')}", file=stream)
        print(f"Length of residual bitstream: {size} bits, "
              f"bytes: {np.ceil(size / 8):.0f} aligned ({np.ceil(size / 8) / (2 ** 20):.2f} MiB)", file=stream)
        lpc_bytes = np.ceil(len(self._data) * self._lpc_precision * self._lpc_order * 1 / 8)
        print(f"Bytes needed for coefficients: {lpc_bytes:.0f} B", file=stream)
        print(f"Output file size: {output_file.stat().st_size} ({output_file.stat().st_size / (2 ** 20):.2f} MiB)",
              file=stream)
        print(f"Grand Ratio = {output_file.stat().st_size / self._source_size:.4f}", file=stream)

        print(f"Size of the resulting dataframe: {self.usage_mib():.3f} MiB", file=stream)



import argparse
from pathlib import Path

from straw.static import Default
from straw.straw import run

"""
The main executable script
When called parses the command line arguments and acts accordingly
"""

parser = argparse.ArgumentParser(description="Lossless multi-channel audio codec")
parser.add_argument("-i", "--input", dest="input_files", metavar="INPUT_FILE", type=str, nargs="+",
                    help="Input files", required=True)
parser.add_argument("-o", "--output", dest="output_file", metavar="OUTPUT_FILE", type=str,
                    help="Output file")
parser.add_argument("-d", "--decode", dest="decode", action="store_true",
                    help="Decode")

parser.add_argument("--figures", dest="figures", action="store_true",
                    help="Generate figures - temporary option for runs with non-consistent behavior")
parser.add_argument("--no-dynamic-blocks", dest="dynamic_blocksize", action="store_false",
                    help="Use dynamic blocksize, by default delimited by short-term energy")
parser.add_argument("--figures-show", dest="fig_show", action="store_true", help="Show figures")
parser.add_argument("--figures-dir", dest="fig_dir", metavar="FIG_DIR", type=str,
                    default="outputs", help="Override figure directory (default='outputs')")

parser.add_argument("--min-frame-size", dest="min_frame_size", metavar="MIN_FRAME_SIZE", type=int,
                    default=Default.min_frame_size, help="Override minimal frame size")
parser.add_argument("--max-frame-size", dest="max_frame_size", metavar="MAX_FRAME_SIZE", type=int,
                    default=Default.max_frame_size, help="Override maximal frame size")
parser.add_argument("--framing-threshold", dest="framing_treshold", metavar="THRESHOLD", type=int,
                    default=Default.framing_treshold, help="Override framing treshold")
parser.add_argument("--framing-resolution", dest="framing_resolution", metavar="RESOLUTION", type=int,
                    default=Default.framing_resolution, help="Override framing resolution")
parser.add_argument("--rice-responsiveness", dest="rice_responsiveness", metavar="RESPONSIVENESS", type=int,
                    default=Default.rice_responsiveness, help="Override Rice codding responsiveness")

parser.add_argument("--no-parallel", dest="parallel", action="store_false",
                    help="Disable parallelization")
parser.add_argument("--silent", dest="silent", action="store_true",
                    help="Silence the coder completely")
parser.add_argument("--verbose", dest="verbose", action="store_true",
                    help="Verbose output")

args = parser.parse_args()

if args.min_frame_size > args.max_frame_size:
    raise ValueError("Minimal frame size can't be larger than maximal frame size")

# Fix args
if args.output_file:
    args.output_file = Path(args.output_file)
else:
    pth = Path(args.input_files[0])
    if args.decode:
        args.output_file = (pth.parent / pth.stem).with_suffix(".wav")
    else:
        args.output_file = (pth.parent / pth.stem).with_suffix(".straw")

if args.figures:
    from figures import plot_all

    plot_all(args)
else:
    run(args)



import math

import numpy as np
import pandas as pd
from numpy.lib.polynomial import roots
from scipy.linalg import solve_toeplitz
from scipy.signal import get_window

from . import ext_lpc


####################
# LPC coefficients #
####################


def _autocorr(signal: np.array, target_len: int) -> np.array:
    """
    Computes autocorrelation
    :param signal: signal to be autocorrelated
    :param target_len: number of coefficients
    :return: autocorrelation coefficients
    """
    return np.asarray([signal[:len(signal) - i].dot(signal[i:]) for i in range(target_len)])


def compute_lpc(signal: np.array, p: int) -> np.array:
    """
    Calculates p LPC coefficients
    For fast Levinson-Durbin implementation see:
    https://docs.scipy.org/doc/scipy/reference/generated/scipy.linalg.solve_toeplitz.html
    :param signal: input signal
    :param p: LPC order
    :return: array of length p containing LPC coefficients
    """

    if isinstance(signal, pd.Series):
        # we are dealing with a multichannel LPC
        # Extend to 64 bits to prevent overflows
        if not signal.apply(np.any).all():
            return None

        window = get_window("tukey", signal[signal.index[0]].shape[0])
        r = np.asarray([_autocorr((s.astype(float) / (1 << 15)) * window, p + 1) for s in signal])
        r = np.mean(r, axis=0)
        # lpc_c = np.zeros(p)
        # for i in range(r.shape[0]):
        #     lpc_c += solve_toeplitz(r[i, :-1], r[i, 1:], check_finite=False)
        # return lpc_c / r.shape[0]
    else:
        if not signal.any():
            return None

        window = get_window("tukey", signal.shape[0])
        r = _autocorr((signal.astype(float) / (1 << 15)) * window, p + 1)

    return solve_toeplitz(r[:-1], r[1:], check_finite=False)


def lpc_is_stable(lpc_c) -> bool:
    """
    Determines whether an LPC filter is stable
    :param lpc_c: LPC coefficients
    :return: True if stable, False otherwise
    """
    poly = np.zeros(lpc_c.shape[0] + 1)
    poly[0] = 1
    poly[1:] = -lpc_c
    return np.max(np.abs(roots(poly))) < 1.0


################
# Quantization #
################


def quant_alt(lpc_c, precision) -> (np.array, int, int):
    """
    Alternative implementation for direct quantization
    More time-efficient but not precise
    Currenty not used
    :param lpc_c: numpy array of LPC coefficients to be quantized
    :param precision: target precition in bits
    :return: tuple(QLP, precision, shift)
    """
    # drop 1 bit for sign
    precision -= 1

    # set limits
    qmax = 1 << precision
    qmin = -qmax
    qmax -= 1

    # calculate qlp
    cmax = np.max(np.abs(lpc_c))
    shift = precision - math.frexp(cmax)[1]
    qlp = (lpc_c * (1 << shift)).round().astype(np.int32)

    # Limit the quantized values
    np.clip(qlp, qmin, qmax, out=qlp)

    return qlp.astype(np.int32), precision, shift


def quantize_lpc_cython(lpc_c, precision) -> (np.array, int, int):
    """
    Wrapper around Cython extension for LPC coefficient quantization
    :param lpc_c: numpy array of LPC coefficients to be quantized
    :param precision: target precition in bits
    :return: tuple(QLP, precision, shift)
    """
    shift = ext_lpc.quantize_lpc(lpc_c, precision)
    return lpc_c.astype(np.int32), precision, shift


def simple_quantize(lpc_c, precision) -> (np.array, int, int):
    """
    Yet another implementation for quantization
    Currenty not used
    :param lpc_c: numpy array of LPC coefficients to be quantized
    :param precision: target precition in bits
    :return: tuple(QLP, precision, shift)
    """

    qmax = 1 << precision - 1
    qmin = -qmax
    qmax -= 1

    for shift in reversed(range(0, precision)):
        vals = (lpc_c * (1 << shift)).astype(np.int32)
        if vals.max() <= qmax and vals.min() >= qmin:
            return vals, precision, shift


##############
# Prediction #
##############


def predict_signal(frame: np.array, qlp: np.array, shift):
    """
    Performs LPC prediction
    The resulting predicted signal starts with the order-th sample
    :param frame: signal frame
    :param qlp: quantized LPC coefficients
    :param shift: coefficient quantization shift
    :return: predicted frame with shape [order:]
    """
    shift = int(shift)
    if qlp is None or len(qlp) == 0:
        return None

    return np.convolve(frame, qlp, mode="full")[len(qlp) - 1:-len(qlp)] >> shift


def predict_compute_residual(frame: np.array, qlp: np.array, shift: int):
    """
    Executes LPC prediction and returns the residual by subtracting the original frame from the predicted signal
    :param frame: signal frame
    :param qlp: quantized LPC coefficients
    :param shift: coefficient quantization shift
    :return: frame residual with shape [order:]
    """
    shift = int(shift)
    residual = frame.copy()
    ext_lpc.compute_residual(frame, residual, qlp, shift)
    tmp = residual[len(qlp):]
    # predicted = predict_signal(frame, qlp, shift)
    # tmp = (frame[len(qlp):] - predicted).astype(frame.dtype)
    if tmp.var() < frame.var():
        # IDEA: we could use the same memory space but this would prevent us from using the raw signal after
        # frame[len(qlp):] = tmp
        # return frame[len(qlp):]
        return tmp
    else:
        return None


###############
# Restoration #
###############


def restore_signal(residual, qlp, lp_quantization, warmup_samples):
    """
    Restores the original signal given the residual with quantized LPC coefficients
    :param residual: residual signal
    :param qlp: quantized LPC coefficients
    :param lp_quantization: quantization level
    :param warmup_samples: warmup samples (the first order samples from the original signal)
    :return: reconstructed signal as a numpy array
    """
    order = qlp.shape[0]
    if order <= 0:
        return None

    # the slower but clearer version...
    data = np.pad(warmup_samples, (0, len(residual)))

    for i in range(order, len(residual) + order):
        _sum = 0
        for j in range(order):
            _sum += qlp[j] * data[i - j - 1]
        data[i] = residual[i - order] + (_sum >> lp_quantization)

    return data


def restore_signal_cython(frame: np.array, qlp: np.array, lp_quantization: int) -> np.array:
    """
    Restores the original signal given the residual with quantized LPC coefficients
    Wrapper around Cython extension for signal restoration
    :param frame: signal array initialized with the first samples from the original signal and the residual
    :param qlp: quantized LPC coefficients
    :param lp_quantization: quantization shift
    :return: reconstructed signal as a numpy array
    """
    ext_lpc.restore_signal(frame, qlp, lp_quantization)



# distutils: language = c
# cython: language_level=3
import cython

#########################
# Signedness correction #
#########################

cdef long _interleave(long x):
    """
    Implementation of the overlap and interleave scheme from https://en.wikipedia.org/wiki/Golomb_coding
    :param x: signed integer to be remaped
    :return: positive interleaved integer
    """
    if x == 0:
        return 0

    if x > 0:
        return 2 * x
    else:
        return -2 * x - 1

@cython.cdivision(True)
cdef long _deinterleave(long x):
    """
    Reverse of _interleave(short x)
    :param x: positive interleaved integer
    :return: original signed integer
    """
    if x == 0:
        return 0

    if x % 2 == 0:
        return x / 2
    else:
        return (x + 1) / -2

cdef void update_scale(long s, short m, short *scale):
    if s > m:
        scale[0] += 1
    elif s < m:
        scale[0] -= 1
    else:
        scale[0] = 0

############
# Encoding #
############

@cython.cdivision(True)
cdef void _push_bit(unsigned char[:] bits, Py_ssize_t *bit_i, char val):
    if val == 0:
        bit_i[0] += 1
        return

    cdef Py_ssize_t byte_i = bit_i[0] / 8
    cdef char real_bit_i = 7 - (bit_i[0] % 8)
    bit_i[0] += 1
    bits[byte_i] |= 1 << real_bit_i

@cython.cdivision(True)
def encode_frame(unsigned char[:] bits, cython.integral[:] frame, short k, short resp, short adaptive):
    """
    Encodes a whole residual frame and appends it to the end of the given bitstream
    :param bits: bitaray to which the bits will be appended
    :param frame: the frame to be encoded
    :param k: starting rice parameter
    :param resp: rice parameter responsiveness
    :param adaptive: if True do adaptive rice coding by varying the parameter
    :return: None
    """
    cdef short j
    cdef long m, q, s
    cdef Py_ssize_t x_max, i
    cdef Py_ssize_t bit_i = 0
    cdef Py_ssize_t bit_i_max = bits.shape[0] * 8
    x_max = frame.shape[0]
    m = 1 << k

    cdef short scale = 0

    for i in range(x_max):
        s = _interleave(frame[i])

        # Quotient code
        q = s / m

        if bit_i + q + k + 1 >= bit_i_max:
            return -1

        bit_i += q
        _push_bit(bits, &bit_i, 1)

        for j in range(k):
            _push_bit(bits, &bit_i, s >> (k - j - 1) & 1)

        # TODO: feed-forward rice implementation
        if not adaptive:
            continue

        # Update rice param
        if scale > resp:
            scale = 0
            k += 1
            m = 1 << k
            # print("e switched up:\t\t ", i, s, m)
            continue
        if scale < -resp and k > 0:
            scale = 0
            k -= 1
            m = 1 << k
            # print("e switched down:\t ", i, s, m)
            continue
        elif scale < -resp:
            scale = -resp

        update_scale(s, m, &scale)

    return bit_i

############
# Decoding #
############

@cython.cdivision(True)
cdef char _get_bit(const unsigned char[:] bits, Py_ssize_t *bit_i):
    cdef Py_ssize_t byte_i = bit_i[0] / 8
    cdef char real_bit_i = 7 - (bit_i[0] % 8)
    bit_i[0] += 1
    return (bits[byte_i] >> real_bit_i) & 1

@cython.cdivision(True)
def decode_frame(cython.integral[:] frame, const unsigned char[:] bits, short k, short resp, short adaptive, long starting_i = 0):
    """
    Decodes a whole residual frame from the given bitstream
    :param frame: numpy array where the decoded frame should be stored
    :param bits: bitaray from which the frame should be restored
    :param k: starting rice parameter
    :param resp: rice parameter responsiveness
    :param adaptive: if True do adaptive rice coding by varying the parameter
    :return: None
    """
    cdef short j
    cdef long m, q, s
    cdef Py_ssize_t x_max, i
    cdef Py_ssize_t bit_i = starting_i
    x_max = frame.shape[0]
    m = 1 << k

    cdef short scale = 0

    for i in range(x_max):
        q = 0
        while not _get_bit(bits, &bit_i):
            q += 1

        s = m * q

        for j in range(k):
            s |= _get_bit(bits, &bit_i) << (k - j - 1)

        frame[i] = _deinterleave(s)

        if not adaptive:
            continue

        if scale > resp:
            scale = 0
            k += 1
            m = 1 << k
            # print("dec switched up:\t ", i, s, m)
            continue
        if scale < -resp and k > 0:
            scale = 0
            k -= 1
            m = 1 << k
            # print("dec switched down:\t ", i, s, m)
            continue
        elif scale < -resp:
            scale = -resp

        update_scale(s, m, &scale)

    return bit_i - starting_i

###########
# Utility #
###########

@cython.cdivision(True)
def kparams(cython.integral[:] frame, short k, short resp):
    """
    Encodes a whole residual frame and appends it to the end of the given bitstream
    :param frame: the frame to be encoded
    :param k: starting rice parameter
    :param resp: rice parameter responsiveness
    :return: None
    """
    cdef long s
    cdef Py_ssize_t x_max, i
    x_max = frame.shape[0]
    m = 1 << k

    cdef short scale = 0

    for i in range(x_max):
        s = _interleave(frame[i])
        frame[i] = k

        if scale > resp:
            scale = 0
            k += 1
            m = 1 << k
            continue
        if scale < -resp and k > 0:
            scale = 0
            k -= 1
            m = 1 << k
            continue

        update_scale(s, m, &scale)

def interleave_frame(cython.integral[:] frame):
    cdef Py_ssize_t x_max, i
    x_max = frame.shape[0]
    for i in range(x_max):
        frame[i] = _interleave(frame[i])



from pathlib import Path

from setuptools import Extension, setup, find_packages

ROOT = Path(__file__).parent

extensions = [
    Extension("straw.lpc.ext_lpc", ["src/straw/lpc/ext_lpc.pyx"]),
    Extension("straw.rice.ext_rice", ["src/straw/rice/ext_rice.pyx"]),
    Extension("straw.io.ext_io", ["src/straw/io/ext_io.pyx"]),
]

with (ROOT / "requirements.txt").open("r") as f:
    requirements = [
        line.strip()
        for line in f.readlines()
        if not line.startswith("#")
    ]

with (ROOT / "README.md").open("r") as f:
    README = f.read()

setup(
    # This is deprecated and replaced by pyproject.toml
    # setup_requires=[
    #     # Setuptools 18.0 properly handles Cython extensions.
    #     "setuptools>=18.0",
    #     "cython",
    # ],
    name="straw-codec",
    version="0.4",
    description="Lossless codec for multichannel audio",
    author="KLZ-0",
    author_email="adrian@kalazi.com",
    url="https://github.com/KLZ-0/straw/",
    packages=[pkg for pkg in find_packages(ROOT / "src") if "figures" not in pkg],
    install_requires=requirements,
    package_dir={"": "src"},
    ext_modules=extensions,
    python_requires=">=3.6",
    long_description=README,
    long_description_content_type="text/markdown",
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    entry_points={
        "console_scripts": ["straw=straw:__main__"],
    },
)



from bitarray import bitarray
from bitarray.util import ba2int


class SlicedBitarray(bitarray):
    """
    Sliced version of bitarray
    Holds a pointer to the current position within the bitstream
    Querying a value from the bitstream moves the pointer forward unless specified otherwise
    """
    _current_ptr: int = 0

    def get_int(self, length: int = 1, signed=False) -> int:
        """
        Get an integer from the bitstream
        :param length: length of the integer in bits
        :param signed: whether the integer is in a signed format
        :return: integer
        """
        part = self[self._current_ptr:self._current_ptr + length]
        if len(part) == 0:
            raise ValueError("Unexpected EOF")
        self._current_ptr += length
        return ba2int(part, signed=signed)

    def get_bytes(self, length: int = 8) -> bytes:
        """
        Get bytes from the bitstream
        :param length: number of bits to read
        :return: bytes
        """
        part = self[self._current_ptr:self._current_ptr + length]
        if len(part) == 0:
            raise ValueError("Unexpected EOF")
        self._current_ptr += length
        return part.tobytes()

    def get_int_utf8(self) -> int:
        """
        Get UTF-8 encoded integer from the stream
        NOTE: this function reads until it finds a valid UTF-8 encoded integer value
        :return: integer
        """
        c = self.get_bytes()
        while True:
            try:
                return ord(c.decode("utf-8"))
            except UnicodeDecodeError:
                c += self.get_bytes()

    def get_from(self, start: int) -> bitarray:
        """
        Get a slice of this bitarray from a given position until the current one
        :param start: starting position
        :return: bitarray slice
        """
        return self[start:self._current_ptr]

    def get_pos(self) -> int:
        """
        Get the current position within the bitarray
        :return: bitarray position
        """
        return self._current_ptr

    def advance(self, length: int = 8):
        """
        Advance the bitstream by a given amount of bits
        :param length: number of bits to advance
        :return: None
        """
        self._current_ptr += length

    def skip_padding(self):
        """
        Skip padding until the bitarray is byte-aligned
        :return: None
        """
        if self._current_ptr % 8 != 0:
            self._current_ptr = ((self._current_ptr // 8) + 1) * 8

    def is_eof(self) -> bool:
        """
        Check whether the bitarray is at its end and has no bits to read
        :return: True is EOF, False otherwise
        """
        return len(self[self._current_ptr:self._current_ptr + 1]) == 0



from pathlib import Path
from typing import BinaryIO

import pandas as pd

from straw.io.params import StreamParams
from straw.io.straw import StrawFormatWriter, StrawFormatReader


class Formatter:
    """
    Base formatter class
    """
    @staticmethod
    def validate_dataframe(df: pd.DataFrame, params: StreamParams):
        """
        Validate the contents of the dataframe before opening the file stream
        Raise an error on invalid data
        :param df: source dataframe
        :param params: stream params
        :return: None
        """
        return
        # noinspection PyUnreachableCode
        if len(df) == 0:
            raise ValueError("Empty dataframe")
        if params.channels == 0:
            raise ValueError(f"Invalid number of channels: {params.channels}")
        if params.bits_per_sample == 0:
            raise ValueError(f"Invalid bits per sample: {params.bits_per_sample}")

    def save(self, df: pd.DataFrame, params: StreamParams, output_stream: BinaryIO, flac_mode: bool = False):
        """
        Saves the dataframe into a formatted binary file
        :param df: source dataframe
        :param params: stream params
        :param output_stream: target file
        :param flac_mode: if true the output file will be a FLAC decoder compatible file
        :return: None
        """
        self.validate_dataframe(df, params)
        if flac_mode:
            raise NotImplementedError("FLAC mode is no longer supported")
        else:
            fw = StrawFormatWriter(params, output_stream)
            fw.write(df)
            fw.close_stream(params)

    def load(self, input_file: Path, flac_mode: bool = False, show_progress: bool = True) -> (
    pd.DataFrame, StreamParams):
        if flac_mode:
            raise NotImplementedError("FLAC mode is no longer supported")
        else:
            reader = StrawFormatReader(show_progress)

        reader.load(input_file)
        self.validate_dataframe(reader.get_data(), reader.get_params())
        return reader
